"""
This module contains the routine to initialize quantities like
curvature and torsion from the magnetix axis shape.
"""

import logging
import numpy as np
from scipy.interpolate import CubicSpline as spline
from .spectral_diff_matrix import spectral_diff_matrix
from .util import fourier_minimum
from .fourier_interpolation import fourier_interpolation_matrix

#logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

# Define periodic spline interpolant conversion used in several scripts and plotting
def convert_to_spline(self,array):
    sp=spline(np.append(self.phi,2*np.pi/self.nfp+self.phi[0]), np.append(array,array[0]), bc_type='periodic')
    return sp

def init_axis(self):
    """
    Initialize the curvature, torsion, differentiation matrix, etc.
    """
    # Force curvature to be zero if omn using rc
    omn = self.omn
    rc  = self.rc
    rs  = self.rs
    nfp = self.nfp
    # if omn and np.max(np.abs(self.rs))>0:
    #     # if len(rs)>6:
    #     #     rc[6]=-(1 + rs[2] + rs[4] + (rs[2] + 4 * rs[4]) * 4 * nfp * nfp) / (1 + 36 * nfp * nfp)
    #     # elif len(rs)>4:
    #     #     rs[4]=-(1 + rs[2] + 4 * rs[2] * nfp * nfp) / (1 + 16 * nfp * nfp)
    #     # else:
    #     rs[2]= 1 / (1 + 4 * nfp * nfp)
    # else:
    if omn:
        if len(rc)>6:
            rc[6]=-(1 + rc[2] + rc[4] + (rc[2] + 4 * rc[4]) * 4 * nfp * nfp) / (1 + 36 * nfp * nfp)
        elif len(rc)>4:
            rc[4]=-(1 + rc[2] + 4 * rc[2] * nfp * nfp) / (1 + 16 * nfp * nfp)
        else:
            rc[2]=-1 / (1 + 4 * nfp * nfp)

    # Shorthand:
    nphi = self.nphi
    nfp = self.nfp

    phi = self.phi
    d_phi = self.d_phi
    R0 = np.zeros(nphi)
    Z0 = np.zeros(nphi)
    R0p = np.zeros(nphi)
    Z0p = np.zeros(nphi)
    R0pp = np.zeros(nphi)
    Z0pp = np.zeros(nphi)
    R0ppp = np.zeros(nphi)
    Z0ppp = np.zeros(nphi)
    for jn in range(0, self.nfourier):
        n = jn * nfp
        sinangle = np.sin(n * phi)
        cosangle = np.cos(n * phi)
        R0 += self.rc[jn] * cosangle + self.rs[jn] * sinangle
        Z0 += self.zc[jn] * cosangle + self.zs[jn] * sinangle
        R0p += self.rc[jn] * (-n * sinangle) + self.rs[jn] * (n * cosangle)
        Z0p += self.zc[jn] * (-n * sinangle) + self.zs[jn] * (n * cosangle)
        R0pp += self.rc[jn] * (-n * n * cosangle) + self.rs[jn] * (-n * n * sinangle)
        Z0pp += self.zc[jn] * (-n * n * cosangle) + self.zs[jn] * (-n * n * sinangle)
        R0ppp += self.rc[jn] * (n * n * n * sinangle) + self.rs[jn] * (-n * n * n * cosangle)
        Z0ppp += self.zc[jn] * (n * n * n * sinangle) + self.zs[jn] * (-n * n * n * cosangle)

    d_l_d_phi = np.sqrt(R0 * R0 + R0p * R0p + Z0p * Z0p)
    d2_l_d_phi2 = (R0 * R0p + R0p * R0pp + Z0p * Z0pp) / d_l_d_phi
    d3_l_d_phi3 = (R0p * R0p + R0pp * R0pp + Z0pp * Z0pp + R0 * R0pp + R0p * R0ppp + Z0p * Z0ppp - d2_l_d_phi2 * d2_l_d_phi2) / d_l_d_phi
    B0_over_abs_G0 = nphi / np.sum(d_l_d_phi)
    abs_G0_over_B0 = 1 / B0_over_abs_G0
    self.d_l_d_varphi = abs_G0_over_B0
    G0 = self.sG * abs_G0_over_B0 * self.B0

    # For these next arrays, the first dimension is phi, and the 2nd dimension is (R, phi, Z).
    d_r_d_phi_cylindrical = np.array([R0p, R0, Z0p]).transpose()
    d2_r_d_phi2_cylindrical = np.array([R0pp - R0, 2 * R0p, Z0pp]).transpose()
    d3_r_d_phi3_cylindrical = np.array([R0ppp - 3 * R0p, 3 * R0pp - R0, Z0ppp]).transpose()

    tangent_cylindrical = np.zeros((nphi, 3))
    d_tangent_d_l_cylindrical = np.zeros((nphi, 3))
    for j in range(3):
        tangent_cylindrical[:,j] = d_r_d_phi_cylindrical[:,j] / d_l_d_phi
        d_tangent_d_l_cylindrical[:,j] = (-d_r_d_phi_cylindrical[:,j] * d2_l_d_phi2 / d_l_d_phi \
                                          + d2_r_d_phi2_cylindrical[:,j]) / (d_l_d_phi * d_l_d_phi)

    curvature = np.sqrt(d_tangent_d_l_cylindrical[:,0] * d_tangent_d_l_cylindrical[:,0] + \
                        d_tangent_d_l_cylindrical[:,1] * d_tangent_d_l_cylindrical[:,1] + \
                        d_tangent_d_l_cylindrical[:,2] * d_tangent_d_l_cylindrical[:,2])

    axis_length = np.sum(d_l_d_phi) * d_phi * nfp
    rms_curvature = np.sqrt((np.sum(curvature * curvature * d_l_d_phi) * d_phi * nfp) / axis_length)
    mean_of_R = np.sum(R0 * d_l_d_phi) * d_phi * nfp / axis_length
    mean_of_Z = np.sum(Z0 * d_l_d_phi) * d_phi * nfp / axis_length
    standard_deviation_of_R = np.sqrt(np.sum((R0 - mean_of_R) ** 2 * d_l_d_phi) * d_phi * nfp / axis_length)
    standard_deviation_of_Z = np.sqrt(np.sum((Z0 - mean_of_Z) ** 2 * d_l_d_phi) * d_phi * nfp / axis_length)

    normal_cylindrical = np.zeros((nphi, 3))
    for j in range(3):
        normal_cylindrical[:,j] = d_tangent_d_l_cylindrical[:,j] / curvature
    self.normal_cylindrical = normal_cylindrical

    # b = t x n
    binormal_cylindrical = np.zeros((nphi, 3))
    binormal_cylindrical[:,0] = tangent_cylindrical[:,1] * normal_cylindrical[:,2] - tangent_cylindrical[:,2] * normal_cylindrical[:,1]
    binormal_cylindrical[:,1] = tangent_cylindrical[:,2] * normal_cylindrical[:,0] - tangent_cylindrical[:,0] * normal_cylindrical[:,2]
    binormal_cylindrical[:,2] = tangent_cylindrical[:,0] * normal_cylindrical[:,1] - tangent_cylindrical[:,1] * normal_cylindrical[:,0]

    # If looking for omnigenity, use signed Frenet-Serret frame
    sign_curvature_change = np.ones((self.nphi,))
    if self.omn == True:
        nfp_phi_length = int(np.ceil(self.nphi/2))
        sign_curvature_change[nfp_phi_length:2*nfp_phi_length] = (-1)*np.ones((nfp_phi_length-1,))

    curvature = curvature * sign_curvature_change
    for j in range(3):
        normal_cylindrical[:,j]   =   normal_cylindrical[:,j]*sign_curvature_change
        binormal_cylindrical[:,j] = binormal_cylindrical[:,j]*sign_curvature_change

    self._determine_helicity()
    self.N_helicity = - self.helicity * self.nfp
    
    # We use the same sign convention for torsion as the
    # Landreman-Sengupta-Plunk paper, wikipedia, and
    # mathworld.wolfram.com/Torsion.html.  This sign convention is
    # opposite to Garren & Boozer's sign convention!
    torsion_numerator = (d_r_d_phi_cylindrical[:,0] * (d2_r_d_phi2_cylindrical[:,1] * d3_r_d_phi3_cylindrical[:,2] - d2_r_d_phi2_cylindrical[:,2] * d3_r_d_phi3_cylindrical[:,1]) \
                         + d_r_d_phi_cylindrical[:,1] * (d2_r_d_phi2_cylindrical[:,2] * d3_r_d_phi3_cylindrical[:,0] - d2_r_d_phi2_cylindrical[:,0] * d3_r_d_phi3_cylindrical[:,2]) 
                         + d_r_d_phi_cylindrical[:,2] * (d2_r_d_phi2_cylindrical[:,0] * d3_r_d_phi3_cylindrical[:,1] - d2_r_d_phi2_cylindrical[:,1] * d3_r_d_phi3_cylindrical[:,0]))

    torsion_denominator = (d_r_d_phi_cylindrical[:,1] * d2_r_d_phi2_cylindrical[:,2] - d_r_d_phi_cylindrical[:,2] * d2_r_d_phi2_cylindrical[:,1]) ** 2 \
        + (d_r_d_phi_cylindrical[:,2] * d2_r_d_phi2_cylindrical[:,0] - d_r_d_phi_cylindrical[:,0] * d2_r_d_phi2_cylindrical[:,2]) ** 2 \
        + (d_r_d_phi_cylindrical[:,0] * d2_r_d_phi2_cylindrical[:,1] - d_r_d_phi_cylindrical[:,1] * d2_r_d_phi2_cylindrical[:,0]) ** 2

    torsion = torsion_numerator / torsion_denominator

    # self.d_d_phi = spectral_diff_matrix(self.nphi, xmin = phi[0], xmax = phi[0] + 2*np.pi/self.nfp)
    self.d_d_phi = spectral_diff_matrix(self.nphi, xmin = 0, xmax = 2*np.pi/self.nfp)

    self.Bbar = self.spsi * np.mean(self.B0)

    self.interpolateTo0 = fourier_interpolation_matrix(nphi, -self.phi_shift*self.d_phi*self.nfp)
    # Calculate G0 and varphi
    if self.omn == False:
        # In here B0 is assumed to be given as a fourier series in phi
        # varphi starts from 0 and finishes at 2 * np.pi / nfp
        G0 = self.sG * np.sum(self.B0 * d_l_d_phi) / nphi
        self.varphi = np.zeros(nphi)
        d_l_d_phi_spline = self.convert_to_spline(d_l_d_phi)
        d_l_d_phi_from_zero = d_l_d_phi_spline(np.linspace(0,2*np.pi/self.nfp,self.nphi,endpoint=False))
        for j in range(1, nphi):
            # To get toroidal angle on the full mesh, we need d_l_d_phi on the half mesh.
            self.varphi[j] = self.varphi[j-1] + (d_l_d_phi_from_zero[j-1] + d_l_d_phi_from_zero[j])
        self.varphi = self.varphi * (0.5 * d_phi * 2 * np.pi / axis_length)
    else:
        # In here B0 is assumed to be given as a fourier series in varphi
        # Picard iteration is used to find varphi and G0, with varphi periodic
        # but not starting necessarily at 0
        nu = np.zeros((nphi,))
        for j in range(20):
            varphi = phi + nu
            B0 =  np.array(sum([self.B0_vals[i] *np.cos(nfp*i*varphi) for i in range(len(self.B0_vals))]))
            B0 += np.array(sum([self.B0_svals[i]*np.sin(nfp*i*varphi) for i in range(len(self.B0_svals))]))
            abs_G0 = np.sum(B0 * d_l_d_phi) / nphi
            rhs = -1 + d_l_d_phi * B0 / abs_G0
            last_nu = nu
            nu = np.linalg.solve(self.d_d_phi+self.interpolateTo0, rhs)
            norm_change = np.sqrt(sum((nu-last_nu)**2)/nphi)
            logger.debug("  Iteration {}: |change to nu| = {}".format(j, norm_change))
            if norm_change < 1e-13:
                break
        varphi = phi + nu
        self.varphi = varphi
        B0 =  np.array(sum([self.B0_vals[i] *np.cos(nfp*i*varphi) for i in range(len(self.B0_vals))]))
        B0 += np.array(sum([self.B0_svals[i]*np.sin(nfp*i*varphi) for i in range(len(self.B0_svals))]))
        self.B0 = B0
        G0 = self.sG * np.sum(self.B0 * d_l_d_phi) / nphi
        self.d = np.zeros((self.nphi,))
        if not self.d_over_curvature == 0:
            self.d = self.d_over_curvature * curvature
        elif not self.d_over_curvature_cvals == []:
            if np.size(self.d_over_curvature_cvals) == self.nphi:
                self.d = self.d_over_curvature_cvals * curvature
            else:
                self.d = np.array(sum([self.d_over_curvature_cvals[i]*np.cos(nfp*i*varphi) * curvature for i in range(len(self.d_over_curvature_cvals))]))
        self.d -= self.k_second_order_SS * nfp * self.B0_vals[1] * np.sin(nfp * varphi) / B0
        self.d += np.array(sum([self.d_cvals[i]*np.cos(nfp*i*varphi) for i in range(len(self.d_cvals))]))
        self.d += np.array(sum([self.d_svals[i]*np.sin(nfp*i*varphi) for i in range(len(self.d_svals))]))

    self.d_l_d_varphi = self.sG * G0 / self.B0   
    self.d_d_varphi = np.zeros((nphi, nphi))
    for j in range(nphi):
        self.d_d_varphi[j,:] = self.d_d_phi[j,:] * self.sG * G0 / (self.B0[j] * d_l_d_phi[j])

    self.d_bar = self.d / (curvature + 1e-31)

    self.etabar_squared_over_curvature_squared = (self.B0  / self.Bbar) * self.d_bar**2

    # Add all results to self:
    self.phi = phi
    self.d_phi = d_phi
    self.R0 = R0
    self.Z0 = Z0
    self.R0p = R0p
    self.Z0p = Z0p
    self.R0pp = R0pp
    self.Z0pp = Z0pp
    self.R0ppp = R0ppp
    self.Z0ppp = Z0ppp
    self.G0 = G0
    self.d_l_d_phi = d_l_d_phi
    self.d2_l_d_phi2 = d2_l_d_phi2
    self.d3_l_d_phi3 = d3_l_d_phi3
    self.d_r_d_phi_cylindrical = d_r_d_phi_cylindrical
    self.d2_r_d_phi2_cylindrical = d2_r_d_phi2_cylindrical
    self.d3_r_d_phi3_cylindrical = d3_r_d_phi3_cylindrical
    self.axis_length = axis_length
    self.curvature = curvature
    self.torsion = torsion
    self.d_curvature_d_varphi = np.matmul(self.d_d_varphi, curvature)
    self.d_torsion_d_varphi = np.matmul(self.d_d_varphi, torsion)
    self.d_curvature_d_varphi_at_0 = self.d_curvature_d_varphi[0]
    self.d_d_d_varphi_at_0 = np.matmul(self.d_d_varphi, self.d)[0]
    self.sign_curvature_change = sign_curvature_change
    self.min_R0 = fourier_minimum(self.R0)
    self.min_Z0 = fourier_minimum(self.Z0)
    self.tangent_cylindrical = tangent_cylindrical
    self.normal_cylindrical = normal_cylindrical 
    self.binormal_cylindrical = binormal_cylindrical
    self.abs_G0_over_B0 = abs_G0_over_B0

    # The output is not stellarator-symmetric if (1) R0s is nonzero, (2) Z0c is nonzero, or (3) sigma_initial is nonzero
    if self.order == 'r1':
        self.lasym = np.max(np.abs(self.rs))>0 or np.max(np.abs(self.zc))>0 or np.abs(self.sigma0)>0
    else:
        self.lasym = np.max(np.abs(self.rs))>0 or np.max(np.abs(self.zc))>0 or np.abs(self.sigma0)>0 or np.any(np.array(self.B2c_svals) > 0.0) or np.any(np.array(self.B2c_svals) < 0.0)

    # Functions that converts a toroidal angle phi0 on the axis to the axis radial and vertical coordinates
    self.R0_func = self.convert_to_spline(sum([self.rc[i]*np.cos(i*self.nfp*self.phi) +\
                                               self.rs[i]*np.sin(i*self.nfp*self.phi) \
                                              for i in range(len(self.rc))]))
    self.Z0_func = self.convert_to_spline(sum([self.zc[i]*np.cos(i*self.nfp*self.phi) +\
                                               self.zs[i]*np.sin(i*self.nfp*self.phi) \
                                              for i in range(len(self.zs))]))

    # Spline interpolants for the cylindrical components of the Frenet-Serret frame:
    self.normal_R_spline     = self.convert_to_spline(self.normal_cylindrical[:,0])
    self.normal_phi_spline   = self.convert_to_spline(self.normal_cylindrical[:,1])
    self.normal_z_spline     = self.convert_to_spline(self.normal_cylindrical[:,2])
    self.binormal_R_spline   = self.convert_to_spline(self.binormal_cylindrical[:,0])
    self.binormal_phi_spline = self.convert_to_spline(self.binormal_cylindrical[:,1])
    self.binormal_z_spline = self.convert_to_spline(self.binormal_cylindrical[:,2])
    self.tangent_R_spline = self.convert_to_spline(self.tangent_cylindrical[:,0])
    self.tangent_phi_spline = self.convert_to_spline(self.tangent_cylindrical[:,1])
    self.tangent_z_spline = self.convert_to_spline(self.tangent_cylindrical[:,2])

    # Spline interpolant for the magnetic field on-axis as a function of phi (not varphi)
    self.B0_spline = self.convert_to_spline(self.B0)

    # Spline interpolant for nu = varphi-phi
    nu = self.varphi-self.phi
    self.nu_spline = self.convert_to_spline(nu)
    self.nu_spline_of_varphi = spline(np.append(self.varphi,self.varphi[0]+2*np.pi/self.nfp), np.append(self.varphi-self.phi,self.varphi[0]-self.phi[0]), bc_type='periodic')



#!/usr/bin/env python3

import unittest
from matplotlib.pyplot import savefig
import numpy as np
from qsc.qsc import Qsc
import logging
import os, sys

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class NewtonTests(unittest.TestCase):

    def test_paper_III(self):
        """
        Compare with Paper III
        """
        # rc      = [ 1.0, 0.0,-0.2 ]
        # zs      = [ 0.0, 0.0, 0.35]
        # B0_vals = [ 1.0, 0.1 ]
        # d_svals = [ 0.0, 1.08, 0.26, 0.46]
        # delta   = 0.1 * 2*np.pi
        # nphi    = 151
        # stel = Qsc(rc=rc,zs=zs, nfp=1, B0_vals=B0_vals, d_svals=d_svals, nphi=nphi, omn=True, delta=delta)
        stel = Qsc.from_paper("QI")
        print('iota  =', stel.iota)
        print('max elongation  =', stel.max_elongation)
        print('mean elongation =', stel.mean_elongation)

        logger.info('Comparing to matlab from Paper III')
        matlab_results_folder = "matlab_data/"
        abs_filename_matlab_results_folder = os.path.join(os.path.dirname(__file__), matlab_results_folder)
        def compare_field(field, py_field, rtol=1e-12, atol=1e-11):
            data = np.genfromtxt(abs_filename_matlab_results_folder+field+".txt")
            logger.info('max difference in {}: {}'.format(field, np.max(np.abs(data - py_field))))
            np.testing.assert_allclose(data, py_field, rtol=rtol, atol=atol)

        np.testing.assert_allclose(stel.iota, 0.716646377954308)

        compare_field('torsion', stel.torsion)
        compare_field('curvature', stel.curvature * stel.sign_curvature_change)
        compare_field('curvature_alt', stel.curvature * stel.sign_curvature_change)
        compare_field('d_l_d_phi', stel.d_l_d_phi)
        compare_field('d2_l_d_phi2', stel.d2_l_d_phi2)
        compare_field('d_d_varphi', stel.d_d_varphi)
        compare_field('phi', stel.phi)
        compare_field('varphi', stel.varphi)
        compare_field('d', stel.d)
        compare_field('B0', stel.B0)
        compare_field('axis_length', stel.axis_length)
        compare_field('G0', stel.G0)
        compare_field('d_r_d_phi_cylindrical', stel.d_r_d_phi_cylindrical)
        compare_field('d2_r_d_phi2_cylindrical', stel.d2_r_d_phi2_cylindrical)
        compare_field('d3_r_d_phi3_cylindrical', stel.d3_r_d_phi3_cylindrical)
        compare_field('d_phi', stel.d_phi)
        compare_field('alpha_iota', stel.alpha_iota)
        compare_field('alpha_notIota', stel.alpha_notIota-3*np.pi)
        compare_field('d_alpha_iota_d_varphi', stel.d_alpha_iota_d_varphi)
        compare_field('d_alpha_notIota_d_varphi', stel.d_alpha_notIota_d_varphi)
        compare_field('sigma', stel.sigma)
        compare_field('alpha', stel.alpha-3*np.pi)
        # Need to untwist B, X and Y
        compare_field('B1c_over_B0', (stel.B1s * -np.sin(stel.varphi) + stel.B1c * np.cos(stel.varphi)) / stel.B0)
        compare_field('B1s_over_B0', (stel.B1s *  np.cos(stel.varphi) + stel.B1c * np.sin(stel.varphi)) / stel.B0)
        compare_field('X1s', stel.X1s_untwisted)
        compare_field('X1c', stel.X1c_untwisted)
        compare_field('Y1c', stel.Y1c_untwisted)
        compare_field('Y1s', stel.Y1s_untwisted)

    def test_multiple_NFP(self):
        """
        Create multiple NFP configurations
        """
        rc      = [ 1.0,0.0,-0.41599809655680886,0.0,0.08291443961920232,0.0,-0.008906891641686355,0.0,0.0,0.0,0.0,0.0,0.0 ]
        zs      = [ 0.0,0.0,-0.28721210154364263,0.0,0.08425262593215394,0.0,-0.010427621520053335,0.0,-0.0008921610906627226,0.0,-6.357200965811029e-07,0.0,2.7316247301500753e-07 ]
        rs      = [ 0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0 ]
        zc      = [ 0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0 ]
        sigma0  =  0.0
        B0_vals = [ 1.0,0.15824229612567256 ]
        omn_method ='non-zone'
        k_buffer = 3
        p_buffer = 2
        k_second_order_SS   = 0.0
        d_over_curvature   = 0.48654821249917474
        d_svals = [ 0.0,-0.00023993050759319644,1.6644294162908823e-05,0.00012071143120099562,-1.1664837950174757e-05,-2.443821681789672e-05,2.0922298879435957e-06 ]
        delta   = 0.1
        nfp     = 1
        B2s_svals = [ 0.0,0.27368018673599265,-0.20986698715787325,0.048031641735420336,0.07269565329289157,1.3981498114634812e-07,-9.952017662433159e-10 ]
        B2c_cvals = [ -0.0007280714400220894,0.20739775852289746,0.05816363701644946,0.06465766308954603,0.006987357785313118,1.2229700694973357e-07,-3.057497440766065e-09,0.0 ]
        B2s_cvals = [ 0.0,0.0,0.0,0.0,0.0 ]
        B2c_svals = [ 0.0,0.0,0.0,0.0 ]
        p2      =  0.0
        nphi    =  301
        stel    =  Qsc(sigma0 = sigma0, omn_method = omn_method, p_buffer = p_buffer, k_buffer=k_buffer, rs=rs,zc=zc, rc=rc,zs=zs, nfp=nfp, B0_vals=B0_vals, d_svals=d_svals, nphi=nphi, omn=True, delta=delta, B2c_cvals=B2c_cvals, B2s_svals=B2s_svals, p2=p2, order='r3', k_second_order_SS=k_second_order_SS, d_over_curvature=d_over_curvature, B2s_cvals=B2s_cvals, B2c_svals=B2c_svals)
        iota    =  -0.718394753879415

        print('iota  =', stel.iota)
        print('max elongation  =', stel.max_elongation)
        print('mean elongation =', stel.mean_elongation)
        print('N_helicity =', stel.N_helicity)
        # stel.plot_axis()
        # import matplotlib.pyplot as plt
        # plt.figure()
        # plt.plot(stel.alpha_notIota, label='alpha not iota')
        # plt.legend()
        # plt.figure()
        # plt.plot(stel.alpha_iota, label='alpha iota')
        # plt.legend()
        # plt.figure()
        # plt.plot(stel.alpha, label='alpha')
        # plt.plot(stel.alpha_no_buffer, label='alpha no buffer')
        # plt.legend()
        # plt.show()

    def test_non_zone(self):
        rc      = [ 1.0, 0.0,-0.2 ]
        zs      = [ 0.0, 0.0, 0.35]
        B0_vals = [ 1.0, 0.1 ]
        d_svals = [ 0.0, 1.08, 0.26, 0.46]
        delta   = 0.1 * 2*np.pi
        nphi    = 151
        stel = Qsc(omn_method = 'non-zone', rc=rc,zs=zs, nfp=1, B0_vals=B0_vals, d_svals=d_svals, nphi=nphi, omn=True, delta=delta)
        print('iota  =', stel.iota)
        print('max elongation  =', stel.max_elongation)
        print('mean elongation =', stel.mean_elongation)

    def test_omnigenity_from_boozxform(self):
        
        def booz_xform_compare(dic, nphi=200):
            import matplotlib.pyplot as plt
            for name, py in dic.items():
                vmecfile = "/Users/rogeriojorge/local/NearAxis_Optimization/Results/"+name+"/wout_"+name+".nc"
                boozfile = "/Users/rogeriojorge/local/NearAxis_Optimization/Results/"+name+"/boozmn_"+name+".nc"
                order=str(name[-2::])
                stel = Qsc.from_boozxform(booz_xform_file=boozfile, vmec_file=vmecfile, order=order, omn=py.omn, N_phi=nphi, nNormal=-py.N_helicity)
                plt.figure()
                plt.title(name+' B0')
                plt.plot(stel.B0_boozxform_array, 'r-', label='B0 VMEC')
                plt.plot(py.B0,                   'r.', label='B0 Near-Axis')
                plt.legend()
                plt.figure()
                plt.title(name+' B1')
                plt.plot(stel.B1c_boozxform_array, 'r-', label='B1c VMEC')
                plt.plot(py.B1c,                   'r.', label='B1c Near-Axis')
                plt.plot(stel.B1s_boozxform_array, 'b-', label='B1s VMEC')
                plt.plot(py.B1s,                   'b.', label='B1s Near-Axis')
                plt.legend()
                if order!='r1':
                    plt.figure()
                    plt.title(name+' B2')
                    plt.plot(stel.B20_boozxform_array, 'r-', label='B20 VMEC')
                    plt.plot(py.B20,                   'r--', label='B20 Near-Axis')
                    plt.plot(stel.B2c_boozxform_array, 'b-', label='B2c VMEC')
                    plt.plot(py.B2c_array,             'b--', label='B2c Near-Axis')
                    plt.plot(stel.B2s_boozxform_array, 'm-', label='B2s VMEC')
                    plt.plot(py.B2s_array,             'm--', label='B2s Near-Axis')
                    plt.legend()
                plt.show()

        dic  = {}
        nphi = 351

        name   = 'QI_NFP4_r1'
        rc      = [ 1.0,0.0,-0.015384615384615385,0.0,0.0,0.0,0.0 ]
        zs      = [ 0.0,0.0,0.008853685778690713,0.0,0.000246103120755672,0.0,1.241037713848815e-05 ]
        B0_vals = [ 1.0,0.11678284595922511 ]
        d_svals = [ 0.0,1.2943134984231057,0.02025266272263474,-0.09056889724435162 ]
        delta   = 0.7075140861631667
        nfp     = 4
        py      = Qsc(rc=rc,zs=zs, nfp=nfp, B0_vals=B0_vals, d_svals=d_svals, nphi=nphi, omn=True, delta=delta)
        dic[name] = py

        # booz_xform_compare(dic,nphi=nphi)

    def test_B_rms_difference_VMEC(self):
        import booz_xform as bx
        from os import path 
        from subprocess import run
        from scipy.io import netcdf
        import matplotlib.pyplot as plt
        from scipy.optimize import curve_fit
        # Input parameters
        executables_path = "/Users/rogeriojorge/local/NearAxis_Optimization/Executables"
        name="QI"
        r_edge = np.exp(-np.linspace(2,6,6))
        Ntheta = 80
        Nzeta  = 140
        # Theta and Phi Arrays
        theta = np.linspace(0,2*np.pi,Ntheta)
        zeta = np.linspace(0,2*np.pi,Nzeta)
        zeta2D,theta2D = np.meshgrid(zeta,theta)
        # Create pyQSC instance
        # stel=Qsc.from_paper(name)
        # rc      = [ 1.0,0.0,-0.08409539411347328,0.0,0.0075907297331848506,0.0,-0.000439832639503239,0.0,0.0 ]
        # zs      = [ 0.0,0.0,0.054107403635000426,0.0,-0.006677539244689272,0.0,0.0005528215202865916,0.0,-3.6968992508928826e-05 ]
        # rs      = [ 0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0 ]
        # zc      = [ 0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0 ]
        # sigma0  =  0.0
        # B0_vals = [ 1.0,0.17873176948252556 ]
        # omn_method ='non-zone-fourier'
        # k_buffer = 1
        # k_second_order_SS   = 0.0
        # d_over_curvature   = 0.6758786723901443
        # d_svals = [ 0.0,-0.29492716510830536,0.008093286550201534,0.043669675571235564,-0.0011914162939404638 ]
        # delta   = 0.1
        # nfp     = 2
        # nphi    = 251
        # stel    =  Qsc(sigma0 = sigma0, omn_method = omn_method, k_buffer=k_buffer, rs=rs,zc=zc, rc=rc,zs=zs, nfp=nfp, B0_vals=B0_vals, d_svals=d_svals, nphi=nphi, omn=True, delta=delta, d_over_curvature=d_over_curvature, k_second_order_SS=k_second_order_SS)
        rc      = [ 1.0,0.0,-0.4056622889934463,0.0,0.07747378220100756,0.0,-0.007803860877024245,0.0,0.0,0.0,0.0,0.0,0.0 ]
        zs      = [ 0.0,0.0,-0.24769666390049602,0.0,0.06767352436978152,0.0,-0.006980621303449165,0.0,-0.0006816270917189934,0.0,-1.4512784317099981e-05,0.0,-2.839050532138523e-06 ]
        B0_vals = [ 1.0,0.16915531046156507 ]
        omn_method ='non-zone-fourier'
        k_buffer = 3
        k_second_order_SS   = 0.0
        d_over_curvature   = 0.5183783762725197
        d_svals = [ 0.0,0.003563114185517955,0.0002015921485566435,-0.0012178616509882368,-0.00011629450296628697,-8.255825435616736e-07,3.2011540526397e-06 ]
        delta   = 0.1
        nfp     = 1
        nphi    = 251
        stel    =  Qsc(omn_method = omn_method, k_buffer=k_buffer, rc=rc,zs=zs, nfp=nfp, B0_vals=B0_vals, d_svals=d_svals, nphi=nphi, omn=True, delta=delta, d_over_curvature=d_over_curvature, k_second_order_SS=k_second_order_SS)
        # Loop over r_edge
        b_RMS_VMEC_diff = []
        b_RMS_BOOZ_diff = []
        for r in r_edge:
            print('r =',r)
            # Run VMEC
            if not path.exists('wout_'+name+'r'+str(r)+'.nc'):
                # Output to VMEC
                stel.to_vmec('input.'+name+'r'+str(r),r=r,
                params={"ns_array": [16,49],#,101,151],
                        "ftol_array": [1e-13,2e-14],#,1e-13,1e-14],
                        "niter_array": [600,1200],#,1000,2000],
                        "mpol": 14,
                        "ntor": 80
                        }, ntheta=28, ntorMax=80)
                # Run VMEC
                bashCommand = executables_path+"/./xvmec2000 input."+name+'r'+str(r)
                run(bashCommand.split())
                os.remove('mercier.'+name+'r'+str(r))
                # os.remove('input.'+name+'r'+str(r))
                os.remove('threed1.'+name+'r'+str(r))
                os.remove('parvmecinfo.txt')
                os.remove('timings.txt')
                try:
                    os.remove('jxbout_'+name+'r'+str(r)+'.txt')
                except:
                    print('')
            else:
                sys.path.insert(1, "/Users/rogeriojorge/local/NearAxis_Optimization/Plotting")
                # import vmecPlot2
                # vmecPlot2.main('wout_'+name+'r'+str(r)+'.nc',stel,r,savefig=False)
                # plt.show()
                # exit()
            # Run BOOZ_XFORM
            b1 = bx.Booz_xform()
            if not path.exists('boozmn_'+name+'r'+str(r)+'.nc'):
                b1.read_wout('wout_'+name+'r'+str(r)+'.nc')
                b1.compute_surfs = [1,2,4,8,15,25,35,47]#,55,80,99]
                b1.mboz = 150
                b1.nboz = 80
                b1.run()
                b1.write_boozmn('boozmn_'+name+'r'+str(r)+'.nc')
            b1.read_boozmn('boozmn_'+name+'r'+str(r)+'.nc')
            # plt.figure(); bx.surfplot(b1, js=0,  fill=False, ncontours=35)
            # plt.figure(); bx.surfplot(b1, js=3, fill=False, ncontours=35)
            # plt.figure(); bx.surfplot(b1, js=6, fill=False, ncontours=35)
            # plt.figure(); bx.symplot(b1, helical_detail = True, sqrts=True)
            plt.figure(); bx.modeplot(b1, sqrts=True); plt.xlabel(r'$s=\sqrt{\psi/\psi_b}$')
            plt.savefig('BOOZ_modes_r'+str(r)+'.pdf'); plt.close()
            # Get |B| on surface from pyQSC
            bQSC = stel.B_mag(r=r, theta=np.mod(-theta2D+np.pi,2*np.pi), phi=zeta2D, Boozer_toroidal=True, B0=1)
            # Get |B| on surface using VMEC
            f = netcdf.netcdf_file('wout_'+name+'r'+str(r)+'.nc','r',mmap=False)
            xn_nyq = f.variables['xn_nyq'][()]
            xm_nyq = f.variables['xm_nyq'][()]
            ns = f.variables['ns'][()]
            bmnc = f.variables['bmnc'][()]
            lasym = f.variables['lasym__logical__'][()]
            iotaVMEC = f.variables['iotaf'][()][0]
            print('iota QSC =',stel.iota,', iota VMEC =',iotaVMEC)
            if lasym==1:
                bmns = f.variables['bmns'][()]
            else:
                bmns = 0*bmnc
            iradius = ns-1
            bVMEC = np.zeros([Ntheta,Nzeta])
            for imode in range(len(xn_nyq)):
                angle = xm_nyq[imode]*theta2D - xn_nyq[imode]*zeta2D
                bVMEC += bmnc[iradius,imode]*np.cos(angle) + bmns[iradius,imode]*np.sin(angle)
            # Find RMS difference
            bDiff_VMEC = (bVMEC - bQSC)**2
            b_RMS_VMEC_diff.append(np.sqrt(np.sum(bDiff_VMEC) / Nzeta / Ntheta))
            # Get |B| on surface using BOOZ_XFORM
            f = netcdf.netcdf_file('boozmn_'+name+'r'+str(r)+'.nc',mmap=False)
            bmnc = f.variables['bmnc_b'][()]
            xn = f.variables['ixn_b'][()]
            xm = f.variables['ixm_b'][()]
            isurf = bmnc.shape[0]-1
            bBOOZ = np.zeros([Ntheta,Nzeta])
            B0_Booz_temp = np.zeros([Ntheta,Nzeta])
            for imn in range(len(xm)):
                angle = xm[imn] * theta2D - xn[imn] * zeta2D
                bBOOZ = bBOOZ + bmnc[isurf,imn] * np.cos(angle)
                if xm[imn]==0:
                    B0_Booz_temp = B0_Booz_temp + (bmnc[0,imn] * 1.5 - bmnc[1,imn] * 0.5) * np.cos(angle)
            normalization=np.mean(B0_Booz_temp)
            B0_Booz = B0_Booz_temp[0]
            # B0_Booz = normalization * (1 + stel.B0_vals[1] * np.cos(zeta2D))
            B1_Booz = bBOOZ - B0_Booz
            B0vmec,B1cvmec,B1svmec,B20vmec,B2cvmec,B2svmec,iotaVMECt = stel.from_boozxform(
                booz_xform_file='boozmn_'+name+'r'+str(r)+'.nc', max_s_for_fit = 0.8,
                N_phi = stel.nphi, max_n_to_plot = 2, vmec_file='wout_'+name+'r'+str(r)+'.nc',
                nNormal=stel.iota-stel.iotaN, input_stel=stel, show=False, savefig=False)
            plt.figure()
            plt.plot(stel.varphi, B0vmec, label='B0 from pyQSCs from_boozxform')
            plt.plot(zeta, B0_Booz_temp[0], label='B0 from BOOZ in this script')
            plt.plot(zeta, normalization * (1 + stel.B0_vals[1] * np.cos(stel.nfp*zeta2D))[0], label='B0 analytical')
            plt.legend()
            plt.title('r='+str(r))
            plt.savefig('B0_diff_r'+str(r)+'.pdf')
            plt.figure()
            b1QSC = stel.B_mag(r=r, theta=np.mod(-theta2D+np.pi,2*np.pi), phi=zeta2D, Boozer_toroidal=True, B0=0)
            plt.contour(zeta2D,theta2D,b1QSC,10,cmap='plasma', linestyles = 'dashed')
            plt.contour(zeta2D,theta2D,B1_Booz,10,cmap='viridis')
            plt.title('r='+str(r))
            plt.savefig('b1QSC_B1booz_diff_r'+str(r)+'.pdf')
            # Find RMS difference
            bDiff_BOOZ = (bBOOZ - bQSC)**2
            b_RMS_BOOZ_diff.append(np.sqrt(np.sum(bDiff_BOOZ) / Nzeta / Ntheta))
            f.close()
        # Create fitting functions
        aspect_ratio = 1/r_edge
        quadratic_function = lambda x,a: a/x/x

        # Check quadratic scaling
        popt, _ = curve_fit(quadratic_function, aspect_ratio, b_RMS_BOOZ_diff)
        y_quadratic = quadratic_function(aspect_ratio,popt[0])
        plt.figure()
        plt.plot(aspect_ratio, b_RMS_BOOZ_diff, 'r*', label=r'RMS($B_{\mathrm{VMEC}}-B_{\mathrm{construction}}$) [T]')
        plt.plot(aspect_ratio, y_quadratic, 'b-', label=r'$\propto 1/A^2$')
        plt.yscale('log') 
        plt.xscale('log')
        plt.xlabel('Aspect Ratio A')
        plt.legend()
        plt.savefig('RMS_VMEC.png')
        # plt.show()

        # Check quadratic scaling with BOOZ_XFORM
        


if __name__ == "__main__":
    unittest.main()



"""
This module contains the routines to compute
a given flux surface shape at a fixed
off-axis cylindrical toroidal angle
"""

import numpy as np
from scipy.optimize import root_scalar

def Frenet_to_cylindrical_residual_func(phi0, phi_target, qic):
    """
    This function takes a point on the magnetic axis with a given
    toroidal angle phi0, computes the actual toroidal angle phi
    for an associated point at r>0 and finds the difference between
    this phi and the target value of phi

    Args:
        phi0 (float): toroidal angle on the axis
        phi_target (float): standard cylindrical toroidal angle
    """
    sinphi0 = np.sin(phi0)
    cosphi0 = np.cos(phi0)
    R0_at_phi0   = qic.R0_func(phi0)
    X_at_phi0    = qic.X_spline(phi0)
    Y_at_phi0    = qic.Y_spline(phi0)
    normal_R     = qic.normal_R_spline(phi0)
    normal_phi   = qic.normal_phi_spline(phi0)
    binormal_R   = qic.binormal_R_spline(phi0)
    binormal_phi = qic.binormal_phi_spline(phi0)

    normal_x   =   normal_R * cosphi0 -   normal_phi * sinphi0
    normal_y   =   normal_R * sinphi0 +   normal_phi * cosphi0
    binormal_x = binormal_R * cosphi0 - binormal_phi * sinphi0
    binormal_y = binormal_R * sinphi0 + binormal_phi * cosphi0

    total_x = R0_at_phi0 * cosphi0 + X_at_phi0 * normal_x + Y_at_phi0 * binormal_x
    total_y = R0_at_phi0 * sinphi0 + X_at_phi0 * normal_y + Y_at_phi0 * binormal_y

    if qic.order != 'r1':
        Z_at_phi0    = qic.Z_spline(phi0)
        tangent_R    = qic.tangent_R_spline(phi0)
        tangent_phi  = qic.tangent_phi_spline(phi0)

        tangent_x = tangent_R * cosphi0 - tangent_phi * sinphi0
        tangent_y = tangent_R * sinphi0 + tangent_phi * cosphi0

        total_x = total_x + Z_at_phi0 * tangent_x
        total_y = total_y + Z_at_phi0 * tangent_y

    Frenet_to_cylindrical_residual = np.arctan2(total_y, total_x) - phi_target
    # We expect the residual to be less than pi in absolute value, so if it is not, the reason must be the branch cut:
    if (Frenet_to_cylindrical_residual >  np.pi): Frenet_to_cylindrical_residual = Frenet_to_cylindrical_residual - 2*np.pi
    if (Frenet_to_cylindrical_residual < -np.pi): Frenet_to_cylindrical_residual = Frenet_to_cylindrical_residual + 2*np.pi
    return Frenet_to_cylindrical_residual

def Frenet_to_cylindrical_1_point(phi0, qic):
    """
    This function takes a point on the magnetic axis with a given
    toroidal angle phi0 and computes the cylindrical coordinate
    components R and Z for an associated point at r>0

    Args:
        phi0: toroidal angle on the axis
    """
    sinphi0 = np.sin(phi0)
    cosphi0 = np.cos(phi0)
    R0_at_phi0   = qic.R0_func(phi0)
    z0_at_phi0   = qic.Z0_func(phi0)
    X_at_phi0    = qic.X_spline(phi0)
    Y_at_phi0    = qic.Y_spline(phi0)
    Z_at_phi0    = qic.Z_spline(phi0)
    normal_R     = qic.normal_R_spline(phi0)
    normal_phi   = qic.normal_phi_spline(phi0)
    normal_z     = qic.normal_z_spline(phi0)
    binormal_R   = qic.binormal_R_spline(phi0)
    binormal_phi = qic.binormal_phi_spline(phi0)
    binormal_z   = qic.binormal_z_spline(phi0)

    normal_x   =   normal_R * cosphi0 -   normal_phi * sinphi0
    normal_y   =   normal_R * sinphi0 +   normal_phi * cosphi0
    binormal_x = binormal_R * cosphi0 - binormal_phi * sinphi0
    binormal_y = binormal_R * sinphi0 + binormal_phi * cosphi0

    total_x = R0_at_phi0 * cosphi0 + X_at_phi0 * normal_x + Y_at_phi0 * binormal_x
    total_y = R0_at_phi0 * sinphi0 + X_at_phi0 * normal_y + Y_at_phi0 * binormal_y

    total_z = z0_at_phi0           + X_at_phi0 * normal_z + Y_at_phi0 * binormal_z

    if qic.order != 'r1':
        tangent_R   = qic.tangent_R_spline(phi0)
        tangent_phi = qic.tangent_phi_spline(phi0)
        tangent_z   = qic.tangent_z_spline(phi0)

        tangent_x = tangent_R * cosphi0 - tangent_phi * sinphi0
        tangent_y = tangent_R * sinphi0 + tangent_phi * cosphi0

        total_x = total_x + Z_at_phi0 * tangent_x
        total_y = total_y + Z_at_phi0 * tangent_y
        total_z = total_z + Z_at_phi0 * tangent_z

    total_R = np.sqrt(total_x * total_x + total_y * total_y)
    total_phi=np.arctan2(total_y, total_x)

    return total_R, total_z, total_phi

def Frenet_to_cylindrical(self, r, ntheta=20):
    """
    Function to convert the near-axis coordinate system to
    a cylindrical one for a surface at a particular radius,
    outputing the following arrays: R(theta,varphi),
    phi(theta,varphi) and Z(theta,varphi) with R,phi,Z cylindrical
    coordinates and theta and varphi Boozer coordinates

    Args:
        r:  near-axis radius r of the desired boundary surface
        ntheta: resolution in the poloidal angle theta
    """
    nphi_conversion = self.nphi
    theta = np.linspace(0,2*np.pi,ntheta,endpoint=False)
    phi_conversion = np.linspace(0,2*np.pi/self.nfp,nphi_conversion,endpoint=False)
    R_2D = np.zeros((ntheta,nphi_conversion))
    Z_2D = np.zeros((ntheta,nphi_conversion))
    phi0_2D = np.zeros((ntheta,nphi_conversion))
    for j_theta in range(ntheta):
        costheta = np.cos(theta[j_theta])
        sintheta = np.sin(theta[j_theta])
        X_at_this_theta = r * (self.X1c_untwisted * costheta + self.X1s_untwisted * sintheta)
        Y_at_this_theta = r * (self.Y1c_untwisted * costheta + self.Y1s_untwisted * sintheta)
        Z_at_this_theta = 0 * X_at_this_theta
        if self.order != 'r1':
            # We need O(r^2) terms:
            cos2theta = np.cos(2 * theta[j_theta])
            sin2theta = np.sin(2 * theta[j_theta])
            X_at_this_theta += r * r * (self.X20_untwisted + self.X2c_untwisted * cos2theta + self.X2s_untwisted * sin2theta)
            Y_at_this_theta += r * r * (self.Y20_untwisted + self.Y2c_untwisted * cos2theta + self.Y2s_untwisted * sin2theta)
            Z_at_this_theta += r * r * (self.Z20_untwisted + self.Z2c_untwisted * cos2theta + self.Z2s_untwisted * sin2theta)
            if self.order == 'r3':
                # We need O(r^3) terms:
                costheta  = np.cos(theta[j_theta])
                sintheta  = np.sin(theta[j_theta])
                cos3theta = np.cos(3 * theta[j_theta])
                sin3theta = np.sin(3 * theta[j_theta])
                r3 = r * r * r
                X_at_this_theta += r3 * (self.X3c1_untwisted * costheta + self.X3s1_untwisted * sintheta
                                         + self.X3c3_untwisted * cos3theta + self.X3s3_untwisted * sin3theta)
                Y_at_this_theta += r3 * (self.Y3c1_untwisted * costheta + self.Y3s1_untwisted * sintheta
                                         + self.Y3c3_untwisted * cos3theta + self.Y3s3_untwisted * sin3theta)
                Z_at_this_theta += r3 * (self.Z3c1_untwisted * costheta + self.Z3s1_untwisted * sintheta
                                         + self.Z3c3_untwisted * cos3theta + self.Z3s3_untwisted * sin3theta)
        self.X_spline = self.convert_to_spline(X_at_this_theta)
        self.Y_spline = self.convert_to_spline(Y_at_this_theta)
        self.Z_spline = self.convert_to_spline(Z_at_this_theta)
        for j_phi in range(nphi_conversion):
            # Solve for the phi0 such that r0 + X1 n + Y1 b has the desired phi
            phi_target = phi_conversion[j_phi]
            phi0_rootSolve_min = phi_target - 1.0 / self.nfp
            phi0_rootSolve_max = phi_target + 1.0 / self.nfp
            res = root_scalar(Frenet_to_cylindrical_residual_func, xtol=1e-17, rtol=1e-15, maxiter=2000,\
                              args=(phi_target, self), bracket=[phi0_rootSolve_min, phi0_rootSolve_max], x0=phi_target)
            phi0_solution = res.root
            final_R, final_z, _ = Frenet_to_cylindrical_1_point(phi0_solution, self)
            R_2D[j_theta,j_phi] = final_R
            Z_2D[j_theta,j_phi] = final_z
            phi0_2D[j_theta,j_phi] = phi0_solution
            
    return R_2D, Z_2D, phi0_2D


def to_RZ(self,points):
    """
    Function to convert a set of points in (r,theta,phi0) coordinates
    where r=sqrt(2*psi/B0) is the near-axis radius, theta is the
    Boozer poloidal angle and phi0 is the cylindrical angle phi
    on the axis to cylindrical coordinates (R,Z)

    Args:
        points: an array of floats with dimension Nx3 with N the
        number of points to evaluate with each points having
        the (r,theta,phi0) values to evaluate
    """
    R_final = []
    Z_final = []
    Phi_final = []
    for point in points:
        r      = point[0]
        theta  = point[1]
        phi0   = point[2]
        costheta = np.cos(theta)
        sintheta = np.sin(theta)
        X_at_this_theta = r * (self.X1c_untwisted * costheta + self.X1s_untwisted * sintheta)
        Y_at_this_theta = r * (self.Y1c_untwisted * costheta + self.Y1s_untwisted * sintheta)
        Z_at_this_theta = 0 * X_at_this_theta
        if self.order != 'r1':
            # We need O(r^2) terms:
            cos2theta = np.cos(2 * theta)
            sin2theta = np.sin(2 * theta)
            X_at_this_theta += r * r * (self.X20_untwisted + self.X2c_untwisted * cos2theta + self.X2s_untwisted * sin2theta)
            Y_at_this_theta += r * r * (self.Y20_untwisted + self.Y2c_untwisted * cos2theta + self.Y2s_untwisted * sin2theta)
            Z_at_this_theta += r * r * (self.Z20_untwisted + self.Z2c_untwisted * cos2theta + self.Z2s_untwisted * sin2theta)
            if self.order == 'r3':
                # We need O(r^3) terms:
                cos3theta = np.cos(3 * theta)
                sin3theta = np.sin(3 * theta)
                r3 = r * r * r
                X_at_this_theta += r3 * (self.X3c1_untwisted * costheta + self.X3s1_untwisted * sintheta
                                         + self.X3c3_untwisted * cos3theta + self.X3s3_untwisted * sin3theta)
                Y_at_this_theta += r3 * (self.Y3c1_untwisted * costheta + self.Y3s1_untwisted * sintheta
                                         + self.Y3c3_untwisted * cos3theta + self.Y3s3_untwisted * sin3theta)
                Z_at_this_theta += r3 * (self.Z3c1_untwisted * costheta + self.Z3s1_untwisted * sintheta
                                         + self.Z3c3_untwisted * cos3theta + self.Z3s3_untwisted * sin3theta)
        self.X_spline = self.convert_to_spline(X_at_this_theta)
        self.Y_spline = self.convert_to_spline(Y_at_this_theta)
        self.Z_spline = self.convert_to_spline(Z_at_this_theta)
        R, Z, Phi = Frenet_to_cylindrical_1_point(phi0, self)
        R_final.append(R)
        Z_final.append(Z)
        Phi_final.append(Phi)

    return R_final, Z_final, Phi_final



"""
This module contains the functions for solving the sigma equation
and computing diagnostics of the O(r^1) solution.
"""

import logging
import numpy as np
from .util import fourier_minimum
from .newton import newton
from scipy.interpolate import CubicSpline as spline

#logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

def _residual(self, x):
    """
    Residual in the sigma equation, used for Newton's method.  x is
    the state vector, corresponding to sigma on the phi grid,
    except that the first element of x is actually iota.
    """
    sigma = np.copy(x[1::])
    iota = x[0]
    if self.omn == True:
        helicity = - self.helicity
        self.alpha_no_buffer = np.pi*(2*helicity+1/2) + iota * (self.varphi - np.pi/self.nfp)
        self.alpha_notIota = 0 * self.varphi # added varphi term to get proper sized array
        self.alpha_iota = self.varphi - np.pi/self.nfp

        if self.omn_method == 'buffer':
            location_section_I_II   = np.argmin(np.abs(self.varphi-self.delta))
            location_section_II_III = np.argmin(np.abs(self.varphi-(2*np.pi/self.nfp-self.delta)))+1
            varphiI   = self.varphi[0:location_section_I_II]
            varphiIII = self.varphi[location_section_II_III::]

            # Calculate alpha_iota on buffer regions
            alpha_I_II = self.delta - np.pi/self.nfp
            alpha_0 = 0
            alpha_1 = -(2*alpha_0 - 2*alpha_I_II + 1*self.delta) / self.delta
            alpha_3 = 2*(alpha_0 - alpha_I_II + 1*self.delta) / (self.delta * self.delta * self.delta)
            alpha_4 = -(alpha_0 - alpha_I_II + 1*self.delta) / (self.delta * self.delta * self.delta * self.delta)
            self.alpha_iota[0:location_section_I_II]   =   alpha_0 + alpha_1 * varphiI + alpha_3 * (varphiI **3) + alpha_4 * (varphiI **4)
            self.alpha_iota[location_section_II_III::] = -(alpha_0 + alpha_1 * (2*np.pi/self.nfp-varphiIII) + alpha_3 * ((2*np.pi/self.nfp-varphiIII) **3) + alpha_4 * ((2*np.pi/self.nfp-varphiIII) **4))

            alpha_shift = np.pi*(2*helicity+1/2)
            n_for_alpha = helicity

            # Calculate alpha_notIota on buffer regions
            alpha_I_II = 0
            alpha_0 =  - np.pi * n_for_alpha
            alpha_1 = -(2*alpha_0 - 2*alpha_I_II + 0*self.delta) / self.delta
            alpha_3 = 2*(alpha_0 - alpha_I_II + 0*self.delta) / (self.delta * self.delta * self.delta) 
            alpha_4 = -(alpha_0 - alpha_I_II + 0*self.delta) / (self.delta * self.delta * self.delta * self.delta)
            self.alpha_notIota[0:location_section_I_II] =     alpha_0 + alpha_1 * varphiI + alpha_3 * (varphiI **3) + alpha_4 * (varphiI **4)
            self.alpha_notIota[location_section_II_III::] = -(alpha_0 + alpha_1 * (2*np.pi/self.nfp-varphiIII) + alpha_3 * ((2*np.pi/self.nfp-varphiIII) **3) + alpha_4 * ((2*np.pi/self.nfp-varphiIII) **4))
            self.alpha_notIota = self.alpha_notIota + alpha_shift

            self.alpha_at_zero = alpha_shift + alpha_0

            self.d_alpha_iota_d_varphi = np.matmul(self.d_d_varphi, self.alpha_iota)
            self.d_alpha_notIota_d_varphi = n_for_alpha * self.nfp + np.matmul(self.d_d_varphi, self.alpha_notIota - self.varphi * n_for_alpha * self.nfp) # We have to treat the secular part separately here since d_d_varphi assumes periodicity
            
        elif self.omn_method == 'non-zone':
            k = self.k_buffer
            self.alpha_iota += -(np.pi / self.nfp) * (self.varphi*self.nfp/np.pi - 1)**(2*k+1)
            self.alpha_notIota += np.pi*(2*helicity + 1/2 + helicity * (self.varphi*self.nfp/np.pi - 1)**(2*k+1))

            self.d_alpha_iota_d_varphi = 1 - (np.pi / self.nfp) * (2*k+1) * (self.nfp/np.pi) * (self.varphi*self.nfp/np.pi - 1)**(2*k)
            self.d_alpha_notIota_d_varphi = np.pi * helicity * (2*k+1) * (self.nfp/np.pi) * (self.varphi*self.nfp/np.pi - 1)**(2*k)

        elif self.omn_method == 'non-zone-smoother':
            k = self.k_buffer
            p = self.p_buffer
            nu = (2*k+1)*(2*k)/((2*p+1)*(2*p))
            m = helicity
            n = self.nfp
            a_not_iota =  ((np.pi * m) * (np.pi/n)**(-2*k-1)) * (1/(1-nu))
            a_iota     = -((np.pi / n) * (np.pi/n)**(-2*k-1)) * (1/(1-nu))
            b_not_iota = -((np.pi * m) * (np.pi/n)**(-2*p-1)) * (nu/(1-nu))
            b_iota     =  ((np.pi / n) * (np.pi/n)**(-2*p-1)) * (nu/(1-nu))
            self.alpha_iota    += a_iota     * (self.varphi-np.pi/self.nfp)**(2*k+1) + b_iota     * (self.varphi-np.pi/self.nfp)**(2*p+1)
            self.alpha_notIota += a_not_iota * (self.varphi-np.pi/self.nfp)**(2*k+1) + b_not_iota * (self.varphi-np.pi/self.nfp)**(2*p+1)
            self.alpha_notIota += np.pi*(2*helicity + 1/2)

            self.d_alpha_iota_d_varphi    = 1 + a_iota     * (2*k+1) * (self.varphi-np.pi/self.nfp)**(2*k) + b_iota     * (2*p+1) * (self.varphi-np.pi/self.nfp)**(2*p)
            self.d_alpha_notIota_d_varphi =     a_not_iota * (2*k+1) * (self.varphi-np.pi/self.nfp)**(2*k) + b_not_iota * (2*p+1) * (self.varphi-np.pi/self.nfp)**(2*p)

        elif self.omn_method == 'non-zone-fourier':
            x = self.varphi
            Pi = np.pi
            if self.nfp==1:
                if self.k_buffer==1:
                    self.alpha_iota = -(125*(1728*np.sin(x) + 216*np.sin(2*x) + 64*np.sin(3*x) + 27*np.sin(4*x)) + 1728*np.sin(5*x))/(18000*np.pi*np.pi)
                    self.alpha_notIota =  -(125*(72*np.pi*np.pi*(np.pi + 2*x) + 1728*np.sin(x) + 216*np.sin(2*x) + 64*np.sin(3*x) + 27*np.sin(4*x)) + 1728*np.sin(5*x))/(18000*np.pi*np.pi)
                elif self.k_buffer==3:
                    self.alpha_iota = (-21*(8*(120 - 20*Pi**2 + Pi**4) + (15 + 2*Pi**2*(-5 + Pi**2))*np.cos(x))*np.sin(x))/(2.*Pi**6) - \
                        (28*(40 - 60*Pi**2 + 27*Pi**4)*np.sin(3*x))/(243.*Pi**6) - \
                        (21*(15 - 40*Pi**2 + 32*Pi**4)*np.sin(4*x))/(512.*Pi**6) - \
                        (84*(24 + 25*Pi**2*(-4 + 5*Pi**2))*np.sin(5*x))/(15625.*Pi**6)
                    self.alpha_notIota =          -Pi/2. - x - (84*(120 - 20*Pi**2 + Pi**4)*np.sin(x))/Pi**6 - \
                        (21*(15 + 2*Pi**2*(-5 + Pi**2))*np.sin(2*x))/(4.*Pi**6) - \
                        (28*(40 - 60*Pi**2 + 27*Pi**4)*np.sin(3*x))/(243.*Pi**6) - \
                        (21*(15 - 40*Pi**2 + 32*Pi**4)*np.sin(4*x))/(512.*Pi**6) - \
                        (84*(24 + 25*Pi**2*(-4 + 5*Pi**2))*np.sin(5*x))/(15625.*Pi**6)
                else: 
                    logging.raiseExceptions("Not implemented yet")
            elif self.nfp==2:
                if self.k_buffer==1:
                    self.alpha_iota = -(6*np.sin(self.nfp*x) + 3*np.sin(2*self.nfp*x)/4 + 2*np.sin(3*self.nfp*x)/9 + 3*np.sin(4*self.nfp*x)/32 + 6*np.sin(5*self.nfp*x)/125) / (np.pi * np.pi)
                    self.alpha_notIota = helicity * self.nfp * x + np.pi/2 * (1 + 2*helicity) - helicity * 2 * self.alpha_iota
                else: 
                    logging.raiseExceptions("Not implemented yet")
            elif self.nfp==3:
                if self.k_buffer==1:
                    self.alpha_iota = -(4*np.sin(self.nfp*x) + np.sin(2*self.nfp*x)/2 + 4*np.sin(3*self.nfp*x)/27 + np.sin(4*self.nfp*x)/16 + 4*np.sin(5*self.nfp*x)/125) / (np.pi * np.pi)
                    self.alpha_notIota = helicity * self.nfp * x + np.pi/2 * (1 + 2*helicity) - helicity * 2 * self.alpha_iota
                else: 
                    logging.raiseExceptions("Not implemented yet")
            else: 
                logging.raiseExceptions("Not implemented yet")
            self.d_alpha_iota_d_varphi = np.matmul(self.d_d_varphi, self.alpha_iota)
            self.d_alpha_notIota_d_varphi = helicity * self.nfp + np.matmul(self.d_d_varphi, self.alpha_notIota - self.varphi * helicity * self.nfp)

        # Calculate alpha
        self.alpha = self.alpha_iota * iota + self.alpha_notIota

        # Calculate gamma
        self.gamma_iota    = 1 - self.d_alpha_iota_d_varphi
        self.gamma_notIota = - self.d_alpha_notIota_d_varphi
        self.gamma         = self.gamma_iota * iota + self.gamma_notIota
    else:
        self.gamma = iota + self.helicity * self.nfp - np.matmul(self.d_d_varphi, self.alpha)
    r = np.matmul(self.d_d_varphi, sigma) + self.gamma * \
        (self.etabar_squared_over_curvature_squared * self.etabar_squared_over_curvature_squared + 1 + sigma * sigma) \
        - 2 * self.etabar_squared_over_curvature_squared * (-self.spsi * self.torsion + self.I2 / self.Bbar) * self.G0 / self.B0
    #logger.debug("_residual called with x={}, r={}".format(x, r))

    sigma_at_0 = self.interpolateTo0 @ sigma
    return np.append(r,sigma_at_0-self.sigma0)

def _jacobian(self, x):
    """
    Compute the Jacobian matrix for solving the sigma equation. x is
    the state vector, corresponding to sigma on the phi grid,
    except that the first element of x is actually iota.
    """
    sigma = np.copy(x[1::])

    # d (Riccati equation) / d sigma:
    jac = np.copy(self.d_d_varphi)
    for j in range(self.nphi):
        jac[j, j] += self.gamma [j] * 2 * sigma[j]

    # d (Riccati equation) / d iota:
    if self.omn == True:
        gamma_iota = self.gamma_iota
    else:
        gamma_iota = 1
    jac = np.append(np.transpose([gamma_iota * (self.etabar_squared_over_curvature_squared * self.etabar_squared_over_curvature_squared + 1 + sigma * sigma)]),jac,axis=1)

    # d (sigma[0]-sigma0) / dsigma:
    jac = np.append(jac,[np.append(0,self.interpolateTo0)],axis=0)

    #logger.debug("_jacobian called with x={}, jac={}".format(x, jac))
    return jac

def solve_sigma_equation(self):
    """
    Solve the sigma equation.
    """
    x0 = np.full(self.nphi+1, self.sigma0)
    x0[0] = 0 # Initial guess for iota
    """
    soln = scipy.optimize.root(self._residual, x0, jac=self._jacobian, method='lm')
    self.iota = soln.x[0]
    self.sigma = np.copy(soln.x)
    self.sigma[0] = self.sigma0
    """
    self.sigma = newton(self._residual, x0, jac=self._jacobian)
    self.iota = self.sigma[0]
    self.iotaN = self.iota + self.helicity * self.nfp
    self.sigma = self.sigma[1::]

def _determine_helicity(self):
    """
    Determine the integer N associated with the type of quasisymmetry
    by counting the number of times the normal vector rotates
    poloidally as you follow the axis around toroidally.
    """
    quadrant = np.zeros(self.nphi + 1)
    for j in range(self.nphi):
        if self.normal_cylindrical[j,0] >= 0:
            if self.normal_cylindrical[j,2] >= 0:
                quadrant[j] = 1
            else:
                quadrant[j] = 4
        else:
            if self.normal_cylindrical[j,2] >= 0:
                quadrant[j] = 2
            else:
                quadrant[j] = 3
    quadrant[self.nphi] = quadrant[0]

    counter = 0
    for j in range(self.nphi):
        if quadrant[j] == 4 and quadrant[j+1] == 1:
            counter += 1
        elif quadrant[j] == 1 and quadrant[j+1] == 4:
            counter -= 1
        else:
            counter += quadrant[j+1] - quadrant[j]

    # It is necessary to flip the sign of axis_helicity in order
    # to maintain "iota_N = iota + axis_helicity" under the parity
    # transformations.
    counter *= self.spsi * self.sG
    self.helicity = counter / 4

def r1_diagnostics(self):
    """
    Compute various properties of the O(r^1) solution, once sigma and
    iota are solved for.
    """
    # Spline interpolant for the first order components of the magnetic field
    # as a function of phi, not varphi
    self.d_spline = self.convert_to_spline(self.d)
    if self.omn == True:
        self.alpha_tilde = self.alpha#-self.N_helicity*self.varphi
        self.cos_alpha_tilde_spline = self.convert_to_spline(np.cos(self.alpha_tilde))
        self.sin_alpha_tilde_spline = self.convert_to_spline(np.sin(self.alpha_tilde))

        angle = self.alpha - (-self.helicity * self.nfp * self.varphi)
        sinangle = np.sin(angle)
        cosangle = np.cos(angle)
        B1sQI    = 0    # B=B0*(1+d cos(theta-alpha)), B1cQI = B0*d
        B1cQI    = self.B0 * self.d
        self.B1c = (B1cQI * cosangle - B1sQI * sinangle)
        self.B1s = (B1sQI * cosangle + B1cQI * sinangle)
    else:
        self.alpha_tilde = self.alpha
        self.cos_alpha_tilde_spline = self.convert_to_spline(np.cos(self.alpha))
        self.sin_alpha_tilde_spline = self.convert_to_spline(np.sin(self.alpha))

    self.B1c_spline = self.convert_to_spline(self.B1c)
    self.B1s_spline = self.convert_to_spline(self.B1s)
    # X = X1c*cos(theta-N*varphi)+X1s*sin(theta-N*varphi)
    # theta=0 -> X(theta=0)=X1c*cos(N*varphi)-X1s*sin(N*varphi)
    # theta=0 -> Y(theta=0)=Y1c*cos(N*varphi)-Y1s*sin(N*varphi)
    self.X1c = self.B1c / (self.curvature * self.B0)# * (self.sign_curvature_change)
    self.X1s = self.B1s / (self.curvature * self.B0)# * (self.sign_curvature_change)
    self.Y1s = self.sG * self.Bbar * self.curvature * ( self.B1c + self.B1s * self.sigma) / (self.B1c * self.B1c + self.B1s * self.B1s)# * (self.sign_curvature_change)# + 1e-30) 
    self.Y1c = self.sG * self.Bbar * self.curvature * (-self.B1s + self.B1c * self.sigma) / (self.B1c * self.B1c + self.B1s * self.B1s)# * (self.sign_curvature_change)# + 1e-30)

    # self.X1c = self.convert_to_spline(self.B1c / (self.curvature * self.B0))(self.phi - self.phi_shift*self.d_phi)
    # self.X1s = self.convert_to_spline(self.B1s / (self.curvature * self.B0))(self.phi - self.phi_shift*self.d_phi)
    # self.Y1s = self.convert_to_spline(self.sG * self.Bbar * self.curvature * ( self.B1c + self.B1s * self.sigma) / (self.B1c * self.B1c + self.B1s * self.B1s))(self.phi - self.phi_shift*self.d_phi)
    # self.Y1c = self.convert_to_spline(self.sG * self.Bbar * self.curvature * (-self.B1s + self.B1c * self.sigma) / (self.B1c * self.B1c + self.B1s * self.B1s))(self.phi - self.phi_shift*self.d_phi)
    
    # If helicity is nonzero, then the original X1s/X1c/Y1s/Y1c variables are defined with respect to a "poloidal" angle that
    # is actually helical, with the theta=0 curve wrapping around the magnetic axis as you follow phi around toroidally. Therefore
    # here we convert to an untwisted poloidal angle, such that the theta=0 curve does not wrap around the axis.
    if self.helicity == 0:
        self.X1s_untwisted = self.X1s
        self.X1c_untwisted = self.X1c
        self.Y1s_untwisted = self.Y1s
        self.Y1c_untwisted = self.Y1c
    else:
        # import matplotlib.pyplot as plt
        # plt.plot(self.varphi)
        # plt.plot(self.phi - self.phi_shift*self.d_phi + self.nu_spline(self.phi - self.phi_shift*self.d_phi))
        # plt.show()
        # exit()
        # angle = -self.helicity * self.nfp * (self.phi - self.phi_shift*self.d_phi + self.nu_spline(self.phi - self.phi_shift*self.d_phi))
        # print(self.phi - self.phi_shift*self.d_phi + self.nu_spline(self.phi - self.phi_shift*self.d_phi) - self.varphi)
        angle = -self.helicity * self.nfp * self.varphi
        sinangle = np.sin(angle)
        cosangle = np.cos(angle)
        self.X1s_untwisted = (self.X1s *   cosangle  + self.X1c * sinangle)# * self.sign_curvature_change
        self.X1c_untwisted = (self.X1s * (-sinangle) + self.X1c * cosangle)# * self.sign_curvature_change
        self.Y1s_untwisted = (self.Y1s *   cosangle  + self.Y1c * sinangle)# * self.sign_curvature_change
        self.Y1c_untwisted = (self.Y1s * (-sinangle) + self.Y1c * cosangle)# * self.sign_curvature_change
    # Use (R,Z) for elongation in the (R,Z) plane,
    # or use (X,Y) for elongation in the plane perpendicular to the magnetic axis.
    p = self.X1s * self.X1s + self.X1c * self.X1c + self.Y1s * self.Y1s + self.Y1c * self.Y1c
    q = self.X1s * self.Y1c - self.X1c * self.Y1s
    self.elongation = (p + np.sqrt(p * p - 4 * q * q)) / (2 * np.abs(q))
    self.mean_elongation = np.sum(self.elongation * self.d_l_d_phi) / np.sum(self.d_l_d_phi)
    index = np.argmax(self.elongation)
    self.max_elongation = -fourier_minimum(-self.elongation)
    # Area of the ellipse in the plane perpendicular to the magnetic axis
    self.ellipse_area = np.pi * self.sG * self.Bbar / self.B0 + 2 * (self.X1c * self.Y1c - self.X1s * self.Y1s)

    self.d_X1c_d_varphi = np.matmul(self.d_d_varphi, self.X1c)
    self.d_X1s_d_varphi = np.matmul(self.d_d_varphi, self.X1s)
    self.d_Y1s_d_varphi = np.matmul(self.d_d_varphi, self.Y1s)
    self.d_Y1c_d_varphi = np.matmul(self.d_d_varphi, self.Y1c)

    self.calculate_grad_B_tensor()





#!/usr/bin/env python3

import unittest
import numpy as np
from qsc.fourier_interpolation import fourier_interpolation

class FourierInterpolationTests(unittest.TestCase):

    def test_constant(self):
        """
        If data is constant, the interpolated values should be equal to that constant
        """
        for ndata in range(2, 10):
            for const in [-0.4, 1.3]:
                data = np.full(ndata, const)
                for nrequest in range(1, 5):
                    xrequest = np.random.rand(nrequest) * 20 - 10
                    y = fourier_interpolation(data, xrequest)
                    #print('data:', data, ', xrequest:', xrequest, ', y:', y)
                    np.testing.assert_allclose(y, np.full(nrequest, const), rtol=1e-10, atol=1e-10)
                
    def test_single_mode(self):
        """
        Verify that we can interpolate a single Fourier mode exactly
        """
        phase = 0.7
        for ndata in range(3, 10):
            xdata = np.linspace(0, 2 * np.pi, ndata, endpoint=False)
            for const in [-0.4, 1.3]:
                for amp in [-0.6, 2.1]:
                    data = const + amp * np.cos(xdata + phase)
                    for nrequest in range(1, 5):
                        xrequest = np.random.rand(nrequest) * 20 - 10
                        y = fourier_interpolation(data, xrequest)
                        #print('data:', data, ', xrequest:', xrequest, ', y:', y)
                        np.testing.assert_allclose(y, const + amp * np.cos(xrequest + phase), rtol=1e-10, atol=1e-10)
                
    def test_multiple_modes(self):
        """
        Verify that we can interpolate a sum of Fourier modes exactly
        """
        for nmodes in range(5):
            # Make sure we have enough sample points to resolve all the modes
            for ndata in range(nmodes * 2 + 1, 20):
                xdata = np.linspace(0, 2 * np.pi, ndata, endpoint=False)
                amps = np.random.rand(nmodes) * 4 - 2
                phases = np.random.rand(nmodes) * 20 - 10
                data = np.zeros(ndata)
                for n in range(nmodes):
                    data += amps[n] * np.cos(n * xdata + phases[n])
        
                for nrequest in range(1, 5):
                    xrequest = np.random.rand(nrequest) * 20 - 10
                    y = fourier_interpolation(data, xrequest)

                    yexpected = np.zeros(nrequest)
                    for n in range(nmodes):
                        yexpected += amps[n] * np.cos(n * xrequest + phases[n])
                    
                    #print('data:', data, ', xrequest:', xrequest, ', y:', y)
                    np.testing.assert_allclose(y, yexpected, rtol=1e-10, atol=1e-10)

    ## ADD TEST FOR FOURIER_INTERPOLATION_MATRIX
    # FROM THE MATLAB SCRIPT
                
if __name__ == "__main__":
    unittest.main()



"""
This module contains the calculation for the O(r^2) solution
"""

import logging
import numpy as np
from .util import mu0
from .optimize_nae import min_geo_qi_consistency

#logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

def calculate_r2(self):
    """
    Compute the O(r^2) quantities.
    """
    logger.debug('Calculating O(r^2) terms')
    # First, some shorthand:
    nphi = self.nphi
    Bbar = self.Bbar
    B0_over_abs_G0 = self.B0 / np.abs(self.G0)
    abs_G0_over_B0 = 1 / B0_over_abs_G0
    lprime = abs_G0_over_B0
    G0_over_Bbar = self.G0 / self.Bbar
    X1c = self.X1c
    X1s = self.X1s
    Y1s = self.Y1s
    Y1c = self.Y1c
    d_d_varphi = self.d_d_varphi
    iota_N = self.iotaN
    iota = self.iota
    curvature = self.curvature
    torsion = self.torsion
    B0 = self.B0
    G0 = self.G0
    I2 = self.I2
    B1c = self.B1c
    B1s = self.B1s
    # B2s = self.B2s
    # B2c = self.B2c
    p2 = self.p2
    sG = self.sG
    I2_over_Bbar = self.I2 / self.Bbar

    if self.omn:
        # If in QI, the magnetic field is specified in terms of varphi instead of phi
        integral_one_over_B0_squared_over_varphi = np.sum(1 / (B0 * B0)) / nphi
        # B2 in QI has an alpha and theta instead of vartheta
        B2cPlunk = np.array(sum([self.B2c_cvals[i]*np.cos(self.nfp*i*self.varphi) for i in range(len(self.B2c_cvals))]))
        B2cPlunk = B2cPlunk + np.array(sum([self.B2c_svals[i]*np.sin(self.nfp*i*self.varphi) for i in range(len(self.B2c_svals))]))
        B2sPlunk = np.array(sum([self.B2s_cvals[i]*np.cos(self.nfp*i*self.varphi) for i in range(len(self.B2s_cvals))]))
        B2sPlunk = B2sPlunk + np.array(sum([self.B2s_svals[i]*np.sin(self.nfp*i*self.varphi) for i in range(len(self.B2s_svals))]))
        # angle = self.alpha - (-self.helicity * self.nfp * self.varphi)
        # self.B2c_array = B2cPlunk * np.cos(2 * angle) - B2sPlunk * np.sin(2 * angle)
        # self.B2s_array = B2sPlunk * np.cos(2 * angle) + B2cPlunk * np.sin(2 * angle)
    else:
        # If not in QI, the magnetic field is specified in terms of phi
        integral_one_over_B0_squared_over_varphi = np.sum(self.d_l_d_phi / (abs(G0) * B0)) / nphi
        B2c = np.array(sum([self.B2c_cvals[i]*np.cos(self.nfp*i*self.phi) for i in range(len(self.B2c_cvals))]))
        B2c = B2c + np.array(sum([self.B2c_svals[i]*np.sin(self.nfp*i*self.phi) for i in range(len(self.B2c_svals))]))
        B2s = np.array(sum([self.B2s_cvals[i]*np.cos(self.nfp*i*self.phi) for i in range(len(self.B2s_cvals))]))
        B2s = B2s + np.array(sum([self.B2s_svals[i]*np.sin(self.nfp*i*self.phi) for i in range(len(self.B2s_svals))]))
        self.B2c_array=B2c
        self.B2s_array=B2s

    G2 = -mu0 * p2 * G0 * integral_one_over_B0_squared_over_varphi - iota * I2

    ## Calculate beta_0
    rhs_beta_0_equation = 2 * mu0 * p2 * G0 / Bbar * (1/(B0 * B0)-integral_one_over_B0_squared_over_varphi)
    if np.all(rhs_beta_0_equation == np.zeros((1,nphi))[0]):
        beta_0 = np.zeros((1,nphi))[0]
    else:
        beta_0 = np.linalg.solve(d_d_varphi, rhs_beta_0_equation)
        beta_0 = beta_0 - beta_0[0] # Fix to be zero at origin

    ## Calculate beta_1
    # beta_1c = 0
    # beta_1s = -4 * spsi * sG * mu0 * p2 * etabar * abs_G0_over_B0 / (iota_N * B0 * B0)
    matrix_beta_1 = np.zeros((2 * nphi, 2 * nphi))
    rhs_beta_1 = np.zeros(2 * nphi)
    for j in range(nphi):
        # Handle the terms involving d beta_1c / d zeta and d beta_1s / d zeta:
        # ----------------------------------------------------------------
        # Equation 1, terms involving beta_1c:
        matrix_beta_1[j, 0:nphi] = d_d_varphi[j, :]
        # Equation 2, terms involving beta_1s:
        matrix_beta_1[j+nphi, nphi:(2*nphi)] = d_d_varphi[j, :]

        # Now handle the terms involving beta_1c and beta_1s without d/dzeta derivatives:
        # ----------------------------------------------------------------
        matrix_beta_1[j, j + nphi] = matrix_beta_1[j, j + nphi] + iota_N
        matrix_beta_1[j + nphi, j       ] = matrix_beta_1[j + nphi, j       ] - iota_N

    rhs_beta_1[0:nphi] = -4 * mu0 * p2 * G0 * B1c / (Bbar * B0 * B0 * B0)
    rhs_beta_1[nphi:2 * nphi] = -4 * mu0 * p2 * G0 * B1s / (Bbar * B0 * B0 * B0)

    solution_beta_1 = np.linalg.solve(matrix_beta_1, rhs_beta_1)
    beta_1c = solution_beta_1[0:nphi]
    beta_1s = solution_beta_1[nphi:2 * nphi]

    if np.abs(iota_N) < 1e-8:
        print('Warning: |iota_N| is very small so O(r^2) solve will be poorly conditioned. iota_N=', iota_N)

    V1 = X1c * X1c + Y1c * Y1c + Y1s * Y1s + X1s * X1s
    V2 = 2 * (Y1s * Y1c + X1s * X1c)
    V3 = X1c * X1c + Y1c * Y1c - Y1s * Y1s - X1s * X1s

    factor = - B0_over_abs_G0 / 8
    Z20 = beta_0 * Bbar / (2 * G0 * B0_over_abs_G0) + factor * np.matmul(d_d_varphi,V1)
    Z2s = factor*(np.matmul(d_d_varphi,V2) - 2 * iota_N * V3)
    Z2c = factor*(np.matmul(d_d_varphi,V3) + 2 * iota_N * V2)

    qs = np.matmul(d_d_varphi,X1s) - iota_N * X1c - Y1s * torsion * abs_G0_over_B0
    qc = np.matmul(d_d_varphi,X1c) + iota_N * X1s - Y1c * torsion * abs_G0_over_B0
    rs = np.matmul(d_d_varphi,Y1s) - iota_N * Y1c + X1s * torsion * abs_G0_over_B0
    rc = np.matmul(d_d_varphi,Y1c) + iota_N * Y1s + X1c * torsion * abs_G0_over_B0

    # X2s = B0_over_abs_G0 * (np.matmul(d_d_varphi,Z2s) - 2*iota_N*Z2c + B0_over_abs_G0 * ( abs_G0_over_B0*abs_G0_over_B0*self.B2s_array/B0 + (qc * qs + rc * rs)/2)) / curvature
    if self.omn:
        # If the length of the input array B2s_svals is equal to nphi, then interpret it as B2s on the phi grid
        if np.size(self.B2s_svals) == self.nphi:
            X2s = self.B2s_svals
        else:
            X2s = np.array(sum([self.B2s_svals[i]*np.sin(self.nfp*i*self.varphi) for i in range(len(self.B2s_svals))]))
        # If the length of the input array B2s_cvals is equal to nphi, then interpret it as B2s on the phi grid
        if np.size(self.B2s_cvals) == self.nphi:
            X2s = X2s + self.B2s_cvals
        else:
            X2s = X2s + np.array(sum([self.B2s_cvals[i]*np.cos(self.nfp*i*self.varphi) for i in range(len(self.B2s_cvals))]))
        # X2s = np.array(sum([self.B2s_cvals[i]*np.cos(self.nfp*i*self.varphi) for i in range(len(self.B2s_cvals))])) \
        #     + np.array(sum([self.B2s_svals[i]*np.sin(self.nfp*i*self.varphi) for i in range(len(self.B2s_svals))]))
        # self.B2s_array = X2s * curvature * B0 - B0 * B0_over_abs_G0 * (np.matmul(d_d_varphi,Z2s) - 2*iota_N*Z2c - B0_over_abs_G0 * ( \
        #     + 3 * G0 * G0 * B1c * B1s / (2 * B0**4) - X1c * X1s / 2 * (curvature * abs_G0_over_B0)**2 - (qc * qs + rc * rs)/2))
        self.B2s_array = 3 * B1c * B1s / (2 * B0) \
                       - B0/(2*G0*G0)*( B0 * B0 * (qc * qs + rc * rs) \
                                      + curvature * G0 * G0 * (curvature * X1c * X1s - 2 * X2s) \
                                      + 2 * B0 * G0 * sG * (np.matmul(d_d_varphi, Z2s) - 2 * Z2c * iota_N) )
    else:
        X2s = B0_over_abs_G0 * (np.matmul(d_d_varphi,Z2s) - 2*iota_N*Z2c - B0_over_abs_G0 * ( -abs_G0_over_B0*abs_G0_over_B0*self.B2s_array/B0 \
            + 3 * G0 * G0 * B1c * B1s / (2 * B0**4) - X1c * X1s / 2 * (curvature * abs_G0_over_B0)**2 - (qc * qs + rc * rs)/2)) / curvature

    # X2c = B0_over_abs_G0 * (np.matmul(d_d_varphi,Z2c) + 2*iota_N*Z2s - B0_over_abs_G0 * (-abs_G0_over_B0*abs_G0_over_B0*self.B2c_array/B0 \
    #        + abs_G0_over_B0*abs_G0_over_B0*etabar*etabar/2 - (qc * qc - qs * qs + rc * rc - rs * rs)/4)) / curvature
    if self.omn:
        # X2c = self.B2c_array
        # If the length of the input array B2c_svals is equal to nphi, then interpret it as B2c on the phi grid
        if np.size(self.B2c_svals) == self.nphi:
            X2c = self.B2c_svals
        else:
            X2c = np.array(sum([self.B2c_svals[i]*np.sin(self.nfp*i*self.varphi) for i in range(len(self.B2c_svals))]))
        # If the length of the input array B2c_cvals is equal to nphi, then interpret it as B2c on the phi grid
        if np.size(self.B2c_cvals) == self.nphi:
            X2c = X2c + self.B2c_cvals
        else:
            X2c = X2c + np.array(sum([self.B2c_cvals[i]*np.cos(self.nfp*i*self.varphi) for i in range(len(self.B2c_cvals))]))
        # X2c = np.array(sum([self.B2c_cvals[i]*np.cos(self.nfp*i*self.varphi) for i in range(len(self.B2c_cvals))])) \
        #     + np.array(sum([self.B2c_svals[i]*np.sin(self.nfp*i*self.varphi) for i in range(len(self.B2c_svals))]))
        # self.B2c_array = X2c * curvature * B0 - B0 * B0_over_abs_G0 * (np.matmul(d_d_varphi,Z2c) + 2*iota_N*Z2s - B0_over_abs_G0 * ( \
        #     + 3 * G0 * G0 * (B1c*B1c-B1s*B1s)/(4*B0**4) - (X1c*X1c - X1s*X1s)/4*(curvature*abs_G0_over_B0)**2 \
        #     - (qc * qc - qs * qs + rc * rc - rs * rs)/4))
        self.B2c_array = (1/(4*B0)) * (3 * B1c * B1c - 3 * B1s * B1s + B0_over_abs_G0 * B0_over_abs_G0 \
                                      * ( B0 * B0 * (-qc * qc + qs * qs - rc * rc + rs * rs) \
                                        + G0 * G0 * curvature * (curvature * (-X1c * X1c + X1s * X1s) + 4 * X2c)
                                        - 4 * B0 * G0 * sG * (np.matmul(d_d_varphi, Z2c) + 2 * Z2s * iota_N)))
    else:
        X2c = B0_over_abs_G0 * (np.matmul(d_d_varphi,Z2c) + 2*iota_N*Z2s - B0_over_abs_G0 * (-abs_G0_over_B0*abs_G0_over_B0*self.B2c_array/B0 \
               + 3 * G0 * G0 * (B1c*B1c-B1s*B1s)/(4*B0**4) - (X1c*X1c - X1s*X1s)/4*(curvature*abs_G0_over_B0)**2 \
               - (qc * qc - qs * qs + rc * rc - rs * rs)/4)) / curvature

    # Y2s_from_X20 = -sG * spsi * curvature * curvature / (etabar * etabar)
    Y2s_from_X20 = -(X1s * Y1c + X1c * Y1s) / (X1c * X1c + X1s * X1s + 1e-30)
    # Y2s_inhomogeneous = sG * spsi * (-curvature/2 + curvature*curvature/(etabar*etabar)*(-X2c + X2s * sigma))
    Y2s_inhomogeneous = 1/(2 * B0 * (X1c * X1c + X1s * X1s+ 1e-30))*(Bbar * curvature * sG * (-X1c * X1c + X1s * X1s)\
        + 2 * B0 * (X1s * X2c * Y1c + X1c * X2s * Y1c - X1c * X2c * Y1s + X1s * X2s * Y1s))
    # Y2s_from_Y20 = 0 * X1s
    Y2s_from_Y20 = 2 * X1c * X1s / (X1c * X1c + X1s * X1s+ 1e-30)

    # Y2c_from_X20 = -sG * spsi * curvature * curvature * sigma / (etabar * etabar)
    Y2c_from_X20 = (-X1c * Y1c + X1s * Y1s) / (X1c * X1c + X1s * X1s+ 1e-30)
    # Y2c_from_Y20 = np.full(nphi, 1)
    Y2c_from_Y20 = (X1c * X1c - X1s * X1s) / (X1c * X1c + X1s * X1s+ 1e-30)
    # Y2c_inhomogeneous = sG * spsi * curvature * curvature / (etabar * etabar) * (X2s + X2c * sigma)
    Y2c_inhomogeneous = (Bbar * curvature * sG * X1c * X1s + B0 * (X1c * X2c * Y1c - X1s * X2s * Y1c\
        + X1s * X2c * Y1s + X1c * X2s * Y1s)) / (B0 * (X1c * X1c + X1s * X1s+ 1e-30))

    # Note: in the fX* and fY* quantities below, I've omitted the
    # contributions from X20 and Y20 to the d/dzeta terms. These
    # contributions are handled later when we assemble the large
    # matrix.

    # fX0_from_X20 = -4 * sG * spsi * abs_G0_over_B0 * (Y2c_from_X20 * Z2s - Y2s_from_X20 * Z2c)
    fX0_from_X20 = -4 * G0_over_Bbar * (Y2c_from_X20 * Z2s - Y2s_from_X20 * Z2c)
    # fX0_from_Y20 = -torsion * abs_G0_over_B0 - 4 * sG * spsi * abs_G0_over_B0 * (Z2s) \
    #     - spsi * I2_over_B0 * (-2) * abs_G0_over_B0
    fX0_from_Y20 = -torsion * abs_G0_over_B0 -4 * G0_over_Bbar * (Y2c_from_Y20 * Z2s - Y2s_from_Y20 * Z2c) \
        - I2_over_Bbar * (-2) * abs_G0_over_B0
    # fX0_inhomogeneous = curvature * abs_G0_over_B0 * Z20 - 4 * sG * spsi * abs_G0_over_B0 * (Y2c_inhomogeneous * Z2s - Y2s_inhomogeneous * Z2c) \
    #     - spsi * I2_over_B0 * (0.5 * curvature * sG * spsi) * abs_G0_over_B0 + beta_1s * abs_G0_over_B0 / 2 * Y1c
    fX0_inhomogeneous = curvature * abs_G0_over_B0 * Z20 - 4 * G0_over_Bbar * (Y2c_inhomogeneous * Z2s - Y2s_inhomogeneous * Z2c) \
        - I2_over_Bbar * 0.5 * curvature * (X1s*Y1s + X1c*Y1c) * abs_G0_over_B0 - 0.5 * beta_0 * curvature * abs_G0_over_B0 * (X1s * Y1c - X1c * Y1s)\
        - 0.5 * abs_G0_over_B0 * (beta_1c * Y1s - beta_1s * Y1c)

    # fXs_from_X20 = -torsion * abs_G0_over_B0 * Y2s_from_X20 - 4 * spsi * sG * abs_G0_over_B0 * (Y2c_from_X20 * Z20) \
    #     - spsi * I2_over_B0 * (- 2 * Y2s_from_X20) * abs_G0_over_B0
    fXs_from_X20 = -torsion * abs_G0_over_B0 * Y2s_from_X20 - 4 * G0_over_Bbar * (Y2c_from_X20 * Z20) \
        - I2_over_Bbar * (- 2 * Y2s_from_X20) * abs_G0_over_B0 - beta_0 * abs_G0_over_B0 * (-2 * Y2c_from_X20)
    # fXs_from_Y20 = - 4 * spsi * sG * abs_G0_over_B0 * (-Z2c + Z20)
    fXs_from_Y20 = -torsion * abs_G0_over_B0 * Y2s_from_Y20 - 4 * G0_over_Bbar * (-Z2c + Y2c_from_Y20 * Z20) \
        - I2_over_Bbar * (- 2 * Y2s_from_Y20) * abs_G0_over_B0 - beta_0 * abs_G0_over_B0 * (-2 * Y2c_from_Y20)
    # fXs_inhomogeneous = np.matmul(d_d_varphi,X2s) - 2 * iota_N * X2c - torsion * abs_G0_over_B0 * Y2s_inhomogeneous + curvature * abs_G0_over_B0 * Z2s \
    #     - 4 * spsi * sG * abs_G0_over_B0 * (Y2c_inhomogeneous * Z20) \
    #     - spsi * I2_over_B0 * (0.5 * curvature * spsi * sG - 2 * Y2s_inhomogeneous) * abs_G0_over_B0 \
    #     - (0.5) * abs_G0_over_B0 * beta_1s * Y1s
    fXs_inhomogeneous = np.matmul(d_d_varphi,X2s) - 2 * iota_N * X2c - torsion * abs_G0_over_B0 * Y2s_inhomogeneous + curvature * abs_G0_over_B0 * Z2s \
        - 4 * G0_over_Bbar * (Y2c_inhomogeneous * Z20) \
        - I2_over_Bbar * (0.5 * curvature * (X1s * Y1c + X1c * Y1s) - 2 * Y2s_inhomogeneous) * abs_G0_over_B0 \
        - beta_0 * abs_G0_over_B0 * (-2 * Y2c_inhomogeneous + 0.5 * curvature * (X1c * Y1c - X1s * Y1s)) \
        - 0.5 * abs_G0_over_B0 * (beta_1s * Y1s - beta_1c * Y1c)

    # fXc_from_X20 = - torsion * abs_G0_over_B0 * Y2c_from_X20 - 4 * spsi * sG * abs_G0_over_B0 * (-Y2s_from_X20 * Z20) \
    #     - spsi * I2_over_B0 * (- 2 * Y2c_from_X20) * abs_G0_over_B0
    fXc_from_X20 =  -torsion * abs_G0_over_B0 * Y2c_from_X20 - 4 * G0_over_Bbar * (-Y2s_from_X20 * Z20) \
        - I2_over_Bbar * (- 2 * Y2c_from_X20) * abs_G0_over_B0 - beta_0 * abs_G0_over_B0 * (2 * Y2s_from_X20)
    # fXc_from_Y20 = - torsion * abs_G0_over_B0 - 4 * spsi * sG * abs_G0_over_B0 * (Z2s) \
    #     - spsi * I2_over_B0 * (-2) * abs_G0_over_B0
    fXc_from_Y20 = -torsion * abs_G0_over_B0 * Y2c_from_Y20 - 4 * G0_over_Bbar * (Z2s - Y2s_from_Y20 * Z20) \
        - I2_over_Bbar * (- 2 * Y2c_from_Y20) * abs_G0_over_B0 - beta_0 * abs_G0_over_B0 * (2 * Y2s_from_Y20)
    # fXc_inhomogeneous = np.matmul(d_d_varphi,X2c) + 2 * iota_N * X2s - torsion * abs_G0_over_B0 * Y2c_inhomogeneous + curvature * abs_G0_over_B0 * Z2c \
    #     - 4 * spsi * sG * abs_G0_over_B0 * (-Y2s_inhomogeneous * Z20) \
    #     - spsi * I2_over_B0 * (0.5 * curvature * sG * spsi - 2 * Y2c_inhomogeneous) * abs_G0_over_B0 \
    #     - (0.5) * abs_G0_over_B0 * beta_1s * Y1c
    fXc_inhomogeneous = np.matmul(d_d_varphi,X2c) + 2 * iota_N * X2s - torsion * abs_G0_over_B0 * Y2c_inhomogeneous + curvature * abs_G0_over_B0 * Z2c \
        - 4 * G0_over_Bbar * (-Y2s_inhomogeneous * Z20) \
        - I2_over_Bbar * (0.5 * curvature * (X1c * Y1c - X1s * Y1s) - 2 * Y2c_inhomogeneous) * abs_G0_over_B0 \
        - beta_0 * abs_G0_over_B0 * (2 * Y2s_inhomogeneous - 0.5 * curvature * (X1c * Y1s + X1s * Y1c)) \
        - 0.5 * abs_G0_over_B0 * (beta_1c * Y1s + beta_1s * Y1c)

    # fY0_from_X20 = torsion * abs_G0_over_B0 - spsi * I2_over_B0 * (2) * abs_G0_over_B0
    fY0_from_X20 = torsion * abs_G0_over_B0 - I2_over_Bbar * (2) * abs_G0_over_B0
    # fY0_from_Y20 = np.zeros(nphi)
    fY0_from_Y20 = np.zeros(nphi)
    # fY0_inhomogeneous = -4 * spsi * sG * abs_G0_over_B0 * (X2s * Z2c - X2c * Z2s) \
    #     - spsi * I2_over_B0 * (-0.5 * curvature * X1c * X1c) * abs_G0_over_B0 - (0.5) * abs_G0_over_B0 * beta_1s * X1c
    fY0_inhomogeneous = -4 * G0_over_Bbar * (X2s * Z2c - X2c * Z2s) \
        - I2_over_Bbar * (-0.5 * curvature * (X1s * X1s + X1c * X1c)) * abs_G0_over_B0 \
        - 0.5 * abs_G0_over_B0 * (beta_1s * X1c - beta_1c * X1s)

    # fYs_from_X20 = -2 * iota_N * Y2c_from_X20 - 4 * spsi * sG * abs_G0_over_B0 * (Z2c)
    fYs_from_X20 = -2 * iota_N * Y2c_from_X20 - 4 * G0_over_Bbar * (Z2c)
    # fYs_from_Y20 = np.full(nphi, -2 * iota_N)
    fYs_from_Y20 = -2 * iota_N * Y2c_from_Y20
    # fYs_inhomogeneous = np.matmul(d_d_varphi,Y2s_inhomogeneous) - 2 * iota_N * Y2c_inhomogeneous + torsion * abs_G0_over_B0 * X2s \
    #     - 4 * spsi * sG * abs_G0_over_B0 * (-X2c * Z20) - 2 * spsi * I2_over_B0 * X2s * abs_G0_over_B0
    fYs_inhomogeneous = np.matmul(d_d_varphi,Y2s_inhomogeneous) - 2 * iota_N * Y2c_inhomogeneous + torsion * abs_G0_over_B0 * X2s \
        - 4 * G0_over_Bbar * (-X2c * Z20) - I2_over_Bbar * (-curvature * X1s * X1c + 2 * X2s) * abs_G0_over_B0 \
        - beta_0 * abs_G0_over_B0 * (2 * X2c + 0.5 * curvature*  (X1s * X1s - X1c * X1c)) \
        - 0.5 * abs_G0_over_B0 * (beta_1c * X1c - beta_1s * X1s)

    # fYc_from_X20 = 2 * iota_N * Y2s_from_X20 - 4 * spsi * sG * abs_G0_over_B0 * (-Z2s)
    fYc_from_X20 = 2 * iota_N * Y2s_from_X20 - 4 * G0_over_Bbar * (-Z2s)
    # fYc_from_Y20 = np.zeros(nphi)
    fYc_from_Y20 = 2 * iota_N * Y2s_from_Y20
    # fYc_inhomogeneous = np.matmul(d_d_varphi,Y2c_inhomogeneous) + 2 * iota_N * Y2s_inhomogeneous + torsion * abs_G0_over_B0 * X2c \
    #     - 4 * spsi * sG * abs_G0_over_B0 * (X2s * Z20) \
    #     - spsi * I2_over_B0 * (-0.5 * curvature * X1c * X1c + 2 * X2c) * abs_G0_over_B0 + 0.5 * abs_G0_over_B0 * beta_1s * X1c
    fYc_inhomogeneous = np.matmul(d_d_varphi,Y2c_inhomogeneous) + 2 * iota_N * Y2s_inhomogeneous + torsion * abs_G0_over_B0 * X2c \
        - 4 * G0_over_Bbar * (X2s * Z20) - I2_over_Bbar * (0.5 * curvature * (X1s * X1s - X1c * X1c) + 2 * X2c) * abs_G0_over_B0 \
        - beta_0 * abs_G0_over_B0 * (-2 * X2s + curvature * X1s * X1c) \
        + 0.5 * abs_G0_over_B0 * (beta_1c * X1s + beta_1s * X1c)

    matrix = np.zeros((2 * nphi, 2 * nphi))
    right_hand_side = np.zeros(2 * nphi)
    for j in range(nphi):
        # Handle the terms involving d X_0 / d zeta and d Y_0 / d zeta:
        # ----------------------------------------------------------------

        # Equation 1, terms involving X0:
        # Contributions arise from -X1s * fX0 + Y1c * fYs - Y1s * fYc.
        # matrix[j, 0:nphi] = Y1c[j] * d_d_varphi[j, :] * Y2s_from_X20 - Y1s[j] * d_d_varphi[j, :] * Y2c_from_X20
        matrix[j, 0:nphi] = (-X1s[j] + Y1c[j] * Y2s_from_X20 - Y1s[j] * Y2c_from_X20) * d_d_varphi[j, :]
  
        # Equation 1, terms involving Y0:
        # Contributions arise from  -Y1s * fY0 + Y1c * fYs - Y1s * fYc.
        # matrix[j, nphi:(2*nphi)] = -2 * Y1s[j] * d_d_varphi[j, :]
        matrix[j, nphi:(2*nphi)] = (-Y1s[j] - Y1s[j] * Y2c_from_Y20 + Y1c[j] * Y2s_from_Y20) * d_d_varphi[j, :]

        # Equation 2, terms involving X0:
        # Contributions arise from -X1c * fX0 + Y1s * fYs + Y1c * fYc
        # matrix[j+nphi, 0:nphi] = -X1c[j] * d_d_varphi[j, :] + Y1s[j] * d_d_varphi[j, :] * Y2s_from_X20 + Y1c[j] * d_d_varphi[j, :] * Y2c_from_X20
        matrix[j+nphi, 0:nphi] = (-X1c[j] + Y1s[j] * Y2s_from_X20 + Y1c[j] * Y2c_from_X20) * d_d_varphi[j, :]

        # Equation 2, terms involving Y0:
        # Contributions arise from -Y1c * fY0 + Y1s * fYs + Y1c * fYc
        matrix[j+nphi, nphi:(2*nphi)] = (-Y1c[j] + Y1s[j] * Y2s_from_Y20 + Y1c[j] * Y2c_from_Y20) * d_d_varphi[j, :]

        # Now handle the terms involving X_0 and Y_0 without d/dzeta derivatives:
        # ----------------------------------------------------------------

        # matrix[j, j       ] = matrix[j, j       ] + X1c[j] * fXs_from_X20[j] - Y1s[j] * fY0_from_X20[j] + Y1c[j] * fYs_from_X20[j] - Y1s[j] * fYc_from_X20[j]
        matrix[j, j       ] = matrix[j, j       ] - X1s[j] * fX0_from_X20[j] + X1c[j] * fXs_from_X20[j] - X1s[j] * fXc_from_X20[j] - Y1s[j] * fY0_from_X20[j] + Y1c[j] * fYs_from_X20[j] - Y1s[j] * fYc_from_X20[j]
        # matrix[j, j + nphi] = matrix[j, j + nphi] + X1c[j] * fXs_from_Y20[j] - Y1s[j] * fY0_from_Y20[j] + Y1c[j] * fYs_from_Y20[j] - Y1s[j] * fYc_from_Y20[j]
        matrix[j, j + nphi] = matrix[j, j + nphi] - X1s[j] * fX0_from_Y20[j] + X1c[j] * fXs_from_Y20[j] - X1s[j] * fXc_from_Y20[j] - Y1s[j] * fY0_from_Y20[j] + Y1c[j] * fYs_from_Y20[j] - Y1s[j] * fYc_from_Y20[j]

        # matrix[j + nphi, j       ] = matrix[j + nphi, j       ] - X1c[j] * fX0_from_X20[j] + X1c[j] * fXc_from_X20[j] - Y1c[j] * fY0_from_X20[j] + Y1s[j] * fYs_from_X20[j] + Y1c[j] * fYc_from_X20[j]
        matrix[j + nphi, j       ] = matrix[j + nphi, j       ] - X1c[j] * fX0_from_X20[j] + X1s[j] * fXs_from_X20[j] + X1c[j] * fXc_from_X20[j] - Y1c[j] * fY0_from_X20[j] + Y1s[j] * fYs_from_X20[j] + Y1c[j] * fYc_from_X20[j]
        # matrix[j + nphi, j + nphi] = matrix[j + nphi, j + nphi] - X1c[j] * fX0_from_Y20[j] + X1c[j] * fXc_from_Y20[j] - Y1c[j] * fY0_from_Y20[j] + Y1s[j] * fYs_from_Y20[j] + Y1c[j] * fYc_from_Y20[j]
        matrix[j + nphi, j + nphi] = matrix[j + nphi, j + nphi] - X1c[j] * fX0_from_Y20[j] + X1s[j] * fXs_from_Y20[j] + X1c[j] * fXc_from_Y20[j] - Y1c[j] * fY0_from_Y20[j] + Y1s[j] * fYs_from_Y20[j] + Y1c[j] * fYc_from_Y20[j]


    right_hand_side[0:nphi] = -(-X1s * fX0_inhomogeneous + X1c * fXs_inhomogeneous - X1s * fXc_inhomogeneous - Y1s * fY0_inhomogeneous + Y1c * fYs_inhomogeneous - Y1s * fYc_inhomogeneous)
    right_hand_side[nphi:2 * nphi] = -(- X1c * fX0_inhomogeneous + X1s * fXs_inhomogeneous + X1c * fXc_inhomogeneous - Y1c * fY0_inhomogeneous + Y1s * fYs_inhomogeneous + Y1c * fYc_inhomogeneous)

    solution = np.linalg.solve(matrix, right_hand_side)
    X20 = solution[0:nphi]
    Y20 = solution[nphi:2 * nphi]

    # Now that we have X20 and Y20 explicitly, we can reconstruct Y2s, Y2c, and B20:
    Y2s = Y2s_inhomogeneous + Y2s_from_X20 * X20 + Y2s_from_Y20 * Y20
    Y2c = Y2c_inhomogeneous + Y2c_from_X20 * X20 + Y2c_from_Y20 * Y20

    # B20 = B0 * (curvature * X20 - B0_over_abs_G0 * np.matmul(d_d_varphi,Z20) + (0.5) * etabar * etabar - mu0 * p2 / (B0 * B0) \
    #             - 0.25 * B0_over_abs_G0 * B0_over_abs_G0 * (qc * qc + qs * qs + rc * rc + rs * rs))
    B20 = B0 * (curvature * X20 - B0_over_abs_G0 * np.matmul(d_d_varphi, Z20)) + (3/(4*B0)) * (B1c*B1c + B1s*B1s)\
        + (B0/G0)*(G2 + iota * I2) - 0.25 * B0 * curvature * curvature * (X1c*X1c + X1s*X1s)\
        - 0.25 * B0_over_abs_G0 * B0_over_abs_G0 * (qc * qc + qs * qs + rc * rc + rs * rs) * B0
 
    d_l_d_phi = self.d_l_d_phi
    normalizer = 1 / np.sum(d_l_d_phi)
    self.B20_mean = np.sum(B20 * d_l_d_phi) * normalizer
    self.B20_anomaly = B20 - self.B20_mean
    self.B20_residual = np.sqrt(np.sum((B20 - self.B20_mean) * (B20 - self.B20_mean) * d_l_d_phi) * normalizer) / B0
    self.B20_variation = np.max(B20) - np.min(B20)
 
    # In QI,   B2=B20+B2cQI*Cos(2*(theta-iota*phi+nu)) + B2sQI*Sin(2*(theta-iota*phi+nu)), nu= iota*phi - self.alpha
    # In here, B2=B20+B2carray*Cos(2*(theta-N*phi))    + B2sarray*Sin(2*(theta-N*phi))
    angle = - self.alpha + (-self.helicity * self.nfp * self.varphi) # = nu-iotaN*phi
    self.B2cQI = self.B2c_array * np.cos(2*angle) - self.B2s_array * np.sin(2*angle)
    self.B2sQI = self.B2s_array * np.cos(2*angle) + self.B2c_array * np.sin(2*angle)
    self.B2cQI_spline = self.convert_to_spline(self.B2cQI)
    self.B2sQI_spline = self.convert_to_spline(self.B2sQI)

    self.G2 = G2

    self.d_X20_d_varphi = np.matmul(d_d_varphi, X20)
    self.d_X2s_d_varphi = np.matmul(d_d_varphi, X2s)
    self.d_X2c_d_varphi = np.matmul(d_d_varphi, X2c)
    self.d_Y20_d_varphi = np.matmul(d_d_varphi, Y20)
    self.d_Y2s_d_varphi = np.matmul(d_d_varphi, Y2s)
    self.d_Y2c_d_varphi = np.matmul(d_d_varphi, Y2c)
    self.d_Z20_d_varphi = np.matmul(d_d_varphi, Z20)
    self.d_Z2s_d_varphi = np.matmul(d_d_varphi, Z2s)
    self.d_Z2c_d_varphi = np.matmul(d_d_varphi, Z2c)
    self.d2_X1s_d_varphi2 = np.matmul(d_d_varphi, self.d_X1s_d_varphi)
    self.d2_X1c_d_varphi2 = np.matmul(d_d_varphi, self.d_X1c_d_varphi)
    self.d2_Y1c_d_varphi2 = np.matmul(d_d_varphi, self.d_Y1c_d_varphi)
    self.d2_Y1s_d_varphi2 = np.matmul(d_d_varphi, self.d_Y1s_d_varphi)

    # Store all important results in self:
    self.V1 = V1
    self.V2 = V2
    self.V3 = V3

    self.X20 = X20
    self.X2s = X2s
    self.X2c = X2c
    self.Y20 = Y20
    self.Y2s = Y2s
    self.Y2c = Y2c
    self.Z20 = Z20
    self.Z2s = Z2s
    self.Z2c = Z2c
    self.beta_0  = beta_0
    self.beta_1s = beta_1s
    self.beta_1c = beta_1c
    self.B20 = B20
    self.B20_spline = self.convert_to_spline(self.B20)
    self.B2c_spline = self.convert_to_spline(self.B2c_array)
    self.B2s_spline = self.convert_to_spline(self.B2s_array)
    # self.B2cQI_spline = self.convert_to_spline(self.B2cQI)
    # self.B2sQI_spline = self.convert_to_spline(self.B2sQI)
    # self.B20_of_varphi = self.B20_spline(self.varphi-self.nu_spline(self.phi))
    if self.omn:
        d_B0_d_varphi = np.matmul(self.d_d_varphi, self.B0)
        d_d_d_varphi = np.matmul(self.d_d_varphi, self.d)
        d_2_B0_d_varphi2 = np.matmul(self.d_d_varphi, d_B0_d_varphi)
        multiplicative_factor = self.d * self.B0 / d_B0_d_varphi / d_B0_d_varphi /4
        multiplicative_factor = multiplicative_factor
        self.B2cQI_factor = multiplicative_factor * (2*d_B0_d_varphi * (self.d*d_B0_d_varphi+self.B0*d_d_d_varphi) - self.B0*self.d*d_2_B0_d_varphi2)
        self.B2cQI_factor_spline = self.convert_to_spline(self.B2cQI_factor)
        self.B20QI_deviation = self.B20_spline(self.phi) - self.B20_spline(-self.phi)   # B20(phi) = B20(-phi)
        self.B2cQI_deviation = self.B2cQI_spline(self.phi) - self.B2cQI_factor_spline(self.phi) # B2c(phi) = (1/4)*(B0^2 d^2 / B0')'
        self.B2sQI_deviation = self.B2sQI_spline(self.phi) + self.B2sQI_spline(-self.phi) # B2s(phi) =-B2s(-phi)
        self.B20QI_deviation_max = max(abs(self.B20QI_deviation))
        self.B2cQI_deviation_max = max(abs(self.B2cQI_deviation))
        self.B2sQI_deviation_max = max(abs(self.B2sQI_deviation))


    # O(r^2) diagnostics:
    self.mercier()
    self.calculate_grad_grad_B_tensor()
    #self.grad_grad_B_inverse_scale_length_vs_varphi = t.grad_grad_B_inverse_scale_length_vs_varphi
    #self.grad_grad_B_inverse_scale_length = t.grad_grad_B_inverse_scale_length
    self.calculate_r_singularity()

    if self.helicity == 0:
        self.X20_untwisted = self.X20
        self.X2s_untwisted = self.X2s
        self.X2c_untwisted = self.X2c
        self.Y20_untwisted = self.Y20
        self.Y2s_untwisted = self.Y2s
        self.Y2c_untwisted = self.Y2c
        self.Z20_untwisted = self.Z20
        self.Z2s_untwisted = self.Z2s
        self.Z2c_untwisted = self.Z2c
    else:
        angle = -self.helicity * self.nfp * self.varphi
        sinangle = np.sin(angle)
        cosangle = np.cos(angle)
        self.X20_untwisted = self.X20
        self.Y20_untwisted = self.Y20
        self.Z20_untwisted = self.Z20
        sinangle = np.sin(2*angle)
        cosangle = np.cos(2*angle)
        self.X2s_untwisted = self.X2s *   cosangle  + self.X2c * sinangle
        self.X2c_untwisted = self.X2s * (-sinangle) + self.X2c * cosangle
        self.Y2s_untwisted = self.Y2s *   cosangle  + self.Y2c * sinangle
        self.Y2c_untwisted = self.Y2s * (-sinangle) + self.Y2c * cosangle
        self.Z2s_untwisted = self.Z2s *   cosangle  + self.Z2c * sinangle
        self.Z2c_untwisted = self.Z2s * (-sinangle) + self.Z2c * cosangle

def construct_qi_r2(self, order = 1, verbose = 0, params = [], method = "BFGS", X2s_in = 0, fun_opt = min_geo_qi_consistency):
    """
    Construct a configuration that is QI to second order around the magnetic well minimum in stellarator symmetry, 
    for a fixed axis and B0. This is a bare minimum, and things may be easily changed)
     - order: the order of the zeroes of curvature, important for the QI constraint
     - verbose: non-zero to print the steps of the optimisation
     - params: if decided to input parameters chosen to satisfy the 2nd order qi (only used if they belong
        to the scope of inputs defined). The default is empty, and the function uses only d_bar degrees of freedom (the number
        is prescribed by the size of the input array in self, provided a minimal length)
     - method: method used for the optimisation
     - X2s_in: degree of freedom at second order in a precise QI to second order
     - fun_opt: by default the search minimises min_geo_qi_consistency, which measures how well the constriants are satisfied
        for QI at second order. This allow us to use other functions where other features may also be sought, like reasonable 
        elongation
    """
    logger.debug('Constructing QI to O(r^2)')

    # Change d_bar parameters to find a consistent O(r^2) QI configuration
    if not params:
        params = []
        num_param = 2*order+2
        if self.d_over_curvature_cvals == []:
            params += ['d_over_curvaturec({})'.format(j) for j in range(num_param)]
            d_over_curvature_cvals = np.zeros(num_param)
            if not self.d_over_curvature == 0:
                d_over_curvature_cvals[0] = self.d_over_curvature
            else: 
                d_over_curvature_cvals[0] = 0.5
            self.d_over_curvature_cvals == d_over_curvature_cvals
            self._set_names()
            self.calculate()
        if not self.d_over_curvature_cvals == []:
            if np.size(self.d_over_curvature_cvals) >= num_param:
                params += ['d_over_curvaturec({})'.format(j) for j in range(np.size(self.d_over_curvature_cvals))]
            else:
                params += ['d_over_curvaturec({})'.format(j) for j in range(num_param)]
                d_over_curvature_cvals = np.zeros(num_param)
                d_over_curvature_cvals[0:np.size(self.d_over_curvature_cvals)] = self.d_over_curvature_cvals
                self.d_over_curvature_cvals == d_over_curvature_cvals
                self._set_names()
                self.calculate()
    else:
        new_params = []
        for ind, label in enumerate(params):
            # Check if parameter exists
            if not(label in self.names):
                print('The label ',label, 'does not exist, and will be ignored.')
            else:
                new_params.append(label)
        params = new_params

    # Optimisation is performed
    self.order = "r1" # To leave unnecessary computations out (the way the code is written it does unnecessary things anyway)
    self.optimise_params(params, fun_opt = fun_opt, scale = 0, method = method, verbose = verbose, maxiter = 1000, maxfev = 1000, extras = order) # Order is passed to the function

    # Find the X2s and X2c necessary
    X2c, X2s = evaluate_X2c_X2s_QI(self, X2s_in)

    # Redefine the configuration (not sure why this is needed; when I try to change X2c and X2s only, and then run stel.calculate() with order 'r2', the solution for alpha is different, any clue?)
    self.__init__(omn_method = self.omn_method, delta=self.delta, p_buffer=self.p_buffer, k_buffer=self.k_buffer, rc=self.rc,zs=self.zs, nfp=self.nfp, B0_vals=self.B0_vals, nphi=self.nphi, omn=True, order='r2', d_over_curvature_cvals=self.d_over_curvature_cvals, B2c_svals=X2c, B2s_cvals=X2s)

    return self.B2cQI_deviation_max




def evaluate_X2c_X2s_QI(self, X2s_in = 0):
    """
    Construct X2c and X2s (inputs for second order construction) consistent with a configuration that is QI to
    second order. This only works if the residual of min_geo_qi_consistency is 0 (or numerically very small)
    """
    # Check condition
    if min_geo_qi_consistency(self)>0.01:
        print('The QI condition at the turning points does not seem to be satisfied, and thus the second order construction will fail!')
    
    # Define necessary ingredients
    d_d_varphi = self.d_d_varphi
    B0 = self.B0
    dB0 = np.matmul(d_d_varphi, B0)
    B0_over_abs_G0 = self.B0 / np.abs(self.G0)
    abs_G0_over_B0 = 1 / B0_over_abs_G0
    dldphi = abs_G0_over_B0
    X1c = self.X1c
    X1s = self.X1s
    Y1s = self.Y1s
    Y1c = self.Y1c
    iota_N = self.iotaN
    torsion = self.torsion
    curvature = self.curvature

    # Compute some second order quantites
    V2 = 2 * (Y1s * Y1c + X1s * X1c)
    V3 = X1c * X1c + Y1c * Y1c - Y1s * Y1s - X1s * X1s

    factor = - B0_over_abs_G0 / 8
    Z2s = factor*(np.matmul(d_d_varphi,V2) - 2 * iota_N * V3)
    dZ2s = np.matmul(d_d_varphi,Z2s)
    Z2c = factor*(np.matmul(d_d_varphi,V3) + 2 * iota_N * V2)
    dZ2c = np.matmul(d_d_varphi,Z2c)

    qs = np.matmul(d_d_varphi,X1s) - iota_N * X1c - Y1s * torsion * abs_G0_over_B0
    qc = np.matmul(d_d_varphi,X1c) + iota_N * X1s - Y1c * torsion * abs_G0_over_B0
    rs = np.matmul(d_d_varphi,Y1s) - iota_N * Y1c + X1s * torsion * abs_G0_over_B0
    rc = np.matmul(d_d_varphi,Y1c) + iota_N * Y1s + X1c * torsion * abs_G0_over_B0

    Tc = B0/dldphi*(dZ2c + 2*iota_N*Z2s + (qc*qc-qs*qs+rc*rc-rs*rs)/4/dldphi)
    Ts = B0/dldphi*(dZ2s - 2*iota_N*Z2c + (qc*qs+rc*rs)/2/dldphi)

    angle = self.alpha - (-self.helicity * self.nfp * self.varphi)
    c2a1 = np.cos(2*angle)
    s2a1 = np.sin(2*angle)

    # Construct the consistent X2c and X2s
    X2c_tilde = (Tc*c2a1 + Ts*s2a1 + B0*B0*np.matmul(d_d_varphi,self.d*self.d/dB0)/4)/B0/curvature
    if X2s_in == 0 or np.size(X2s_in) == self.nphi:
        X2s_tilde = X2s_in  # Needds to be checked for the right even parity
    else:
        X2s_tilde = 0 # For simplicity for the time being

    X2c = X2c_tilde*c2a1 + X2s_tilde*s2a1
    X2s = X2c_tilde*s2a1 - X2s_tilde*c2a1

    return X2c, X2s






"""
This module optimises parameters of a NAE
"""

import logging
import numpy as np
import matplotlib.pyplot as plt
from scipy import integrate as integ
from scipy import optimize
from .util import mu0
from .fourier_interpolation import fourier_interpolation

#logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

def opt_fun_stel(x_iter, stel, x_param_label, fun_opt, info = {'Nfeval':0}, res_history = [],verbose =0, extras =[], x_orig = [], thresh = 1.5):
    """
    Function that is passed to the optimiser which sets some properties of Qic object and
    evaluates a function handle that returns a residual.

    x_iter = arguments of the optimisation.
    stel = Qic object to evaluate things on
    x_param_label = name of the parameters of the Qic object that the x_iter values refer to
        (if x_orig is not empty, then x_iter correspond to scale factors).
    fun_opt = function handle of the form fun_opt(stel, extra) supplied.
    info = inside information from the optimisation loop (defaulted to the number of
         evaluations).
    res_history = external object to which the value of the residual is attached
    verbose = if set to 1, it prints the details of the evaluation
    extras = optional object to pass on to fun_opt
    x_orig = if given, then x_iter are not the arguments of the parameters denoted by 
        x_param_label, but rather x_iter*x_orig is
    thresh = limit on the value of x_iter for x_orig!=0. x_iter must satisfy 1/thresh<x_iter<thresh,
        so must be thresh > 1.
    """
    info['Nfeval'] += 1
    # Set degrees of freedom to input values 
    x_iter_all = stel.get_dofs()
    if len(x_orig): # If x_orig given, interpret x_iter as scales 
        for ind, label in enumerate(x_param_label):
            # Put a threshold to the allowable scale x_iter
            if np.any(x_iter[ind]>thresh) or np.any(x_iter[ind]<1/thresh):
                res = 1000 # Chosen a value arbitrarily
                res_history.append(res)
                return res
            # Construct new string of parameter values
            x_iter_all[stel.names.index(label)] = x_iter[ind]*x_orig[stel.names.index(label)]
    else:
        for ind, label in enumerate(x_param_label):
            x_iter_all[stel.names.index(label)] = x_iter[ind]
    # Construct the new nae solution
    stel.set_dofs(x_iter_all)
    # Evaluate the residual
    res = fun_opt(stel, extras)
    if verbose:
        print(f"{info['Nfeval']} -", x_iter)
        if stel.order == 'r2':
            print(f"\N{GREEK CAPITAL LETTER DELTA}B2c = {stel.B2cQI_deviation_max:.4f},",
                # f"1/rc = {1/stel.r_singularity:.2f},",
                # f"1/L\N{GREEK CAPITAL LETTER DELTA}B = {np.max(stel.inv_L_grad_B):.2f},",
                f"Residual = {res:.4f}")
        else: 
            print(f"Residual = {res:.4f}")
    res_history.append(res) # Attach residual value to history
    return res

def fun(stel, extras):
    # Second order QI quality residual for B2c (simple case as an example)
    res = stel.B2cQI_deviation_max
    return res

def min_geo_qi_consistency(stel, order = 1):
    """
    Function that computes the consistency conditions of a first order construction with 
    second order QI at the points phi=0,pi/nfp, and returns the mismatch.

    stel = the stellarator construction, Qic.
    order = order of the zeroes of curvature at the points where the consistency conditions
        are to be evaluated. This determines the number of constraints.
    
    """
    # Define various quantities obtained from Qic
    d_d_varphi = stel.d_d_varphi
    B0 = stel.B0
    dB0 = np.matmul(d_d_varphi, B0)
    B0_over_abs_G0 = stel.B0 / np.abs(stel.G0)
    abs_G0_over_B0 = 1 / B0_over_abs_G0
    dldphi = abs_G0_over_B0
    X1c = stel.X1c
    X1s = stel.X1s
    Y1s = stel.Y1s
    Y1c = stel.Y1c
    iota_N = stel.iotaN
    torsion = stel.torsion

    # Construct second order quantities (like in calculate_r2), but only needing
    # order = 'r1' quantities
    V2 = 2 * (Y1s * Y1c + X1s * X1c)
    V3 = X1c * X1c + Y1c * Y1c - Y1s * Y1s - X1s * X1s

    factor = - B0_over_abs_G0 / 8
    Z2s = factor*(np.matmul(d_d_varphi,V2) - 2 * iota_N * V3)
    dZ2s = np.matmul(d_d_varphi,Z2s)
    Z2c = factor*(np.matmul(d_d_varphi,V3) + 2 * iota_N * V2)
    dZ2c = np.matmul(d_d_varphi,Z2c)

    qs = np.matmul(d_d_varphi,X1s) - iota_N * X1c - Y1s * torsion * abs_G0_over_B0
    qc = np.matmul(d_d_varphi,X1c) + iota_N * X1s - Y1c * torsion * abs_G0_over_B0
    rs = np.matmul(d_d_varphi,Y1s) - iota_N * Y1c + X1s * torsion * abs_G0_over_B0
    rc = np.matmul(d_d_varphi,Y1c) + iota_N * Y1s + X1c * torsion * abs_G0_over_B0

    # Construct the expressions necessary for the conditions
    Tc = B0/dldphi*(dZ2c + 2*iota_N*Z2s + (qc*qc-qs*qs+rc*rc-rs*rs)/4/dldphi)
    Ts = B0/dldphi*(dZ2s - 2*iota_N*Z2c + (qc*qs+rc*rs)/2/dldphi)

    angle = stel.alpha - (-stel.helicity * stel.nfp * stel.varphi)
    c2a1 = np.cos(2*angle)
    s2a1 = np.sin(2*angle)

    # It is unnecessary to compute this everywhere, but the way implemented this is the 
    # easy way
    cond_1st = (Tc*c2a1 + Ts*s2a1 + B0*B0*np.matmul(d_d_varphi,stel.d*stel.d/dB0)/4)
    pos = [0, np.pi] # Not necessary to impose it at 0, but it makes it easier to construct X2s and X2c later
    # Interpolate to the points required (especially because the original grid is shifted)
    cond_1st_eval= fourier_interpolation(cond_1st, pos-stel.phi_shift*stel.d_phi*stel.nfp)
    res = np.sum(cond_1st_eval*cond_1st_eval)
    der_cond = cond_1st
    # For higher order zeroes, we have additional conditions
    if order and order>1:
        for i in range(order-1):
            der_cond = np.matmul(d_d_varphi,der_cond)
            der_cond_eval = fourier_interpolation(der_cond, pos-stel.phi_shift*stel.d_phi*stel.nfp)
            res_add = np.sum(der_cond_eval*der_cond_eval)
            res += res_add

    # Return the mismatch
    return res

def optimise_params(stel, x_param_label, fun_opt = fun, verbose = 0, maxiter = 200, maxfev  = 200, method = 'Nelder-Mead', scale = 0, extras = [], thresh = 1.5):
    """
    Method that optimises stel wrt x_param_label for minimising the function fun_opt.

    stel = the nae Qic object to optimise
    x_param_label = name of the parameters of the Qic object to optimise. Ex.: (['etabar', 'rc(1)'])
    fun_opt = function handle of the form fun_opt(stel, extra) supplied as the measure of
        the residual.
    verbose = if set to anything different from 0, it prints the details of the optimisation
    maxiter = maximum number of iterations in the optimisation
    maxfev = maximum number of function evaluations in the optimisation
    method = method of the optimisation (possibilities are the methods that can be passed to
        scipy.optimize.minimize)
    scale = if it is set to anything other than 0, then it performs optimisation by considering 
        variations of the initial parameters by no more than 100*(thresh-1 or 1-1/thresh)%
    extras = optional object to pass on to fun_opt
    thresh = limit the value by which the optimised parameters may deviate from their initial 
        value (only works if scale!=0): 100*(thresh-1 or 1-1/thresh)%. It must be thresh>1.
    """
    # Extract initial values for the parameters to optimise
    x0_all = stel.get_dofs()
    x_parameter_label_checked = [] # List of available parameters to change
    x0 = []
    if scale:
        for ind, label in enumerate(x_param_label):
            # Check if parameter exists
            if label in stel.names:
                x0.append(1.0)
                x_parameter_label_checked.append(label)
            else:
                print('The label ',label, 'does not exist, and will be ignored.')
    else:
        for ind, label in enumerate(x_param_label):
            if label in stel.names:
                x0.append(x0_all[stel.names.index(label)])
                x_parameter_label_checked.append(label)
            else:
                print('The label ',label, 'does not exist, and will be ignored.')

    # Initialise the residual history 
    res_history = []
    res_history.append(fun_opt(stel, extras))
    
    # Optimisation (using scipy.optimize.minimize)
    if scale:
        # If scale, then limited optimisation relative to initial values
        opt = optimize.minimize(opt_fun_stel, x0, args=(stel, x_parameter_label_checked, fun_opt, {'Nfeval':0}, res_history, verbose, extras, x0_all, thresh), method=method, tol=1e-3, options={'maxiter': maxiter, 'maxfev': maxfev})
    else:
        opt = optimize.minimize(opt_fun_stel, x0, args=(stel, x_parameter_label_checked, fun_opt, {'Nfeval':0}, res_history, verbose, extras), method=method, tol=1e-3, options={'maxiter': maxiter, 'maxfev': maxfev})

    # Make sure that the final form of stel is with the optimised paramters
    x = stel.get_dofs()
    if scale:
        for ind, label in enumerate(x_parameter_label_checked):
            x[stel.names.index(label)] = opt.x[ind]*x0_all[stel.names.index(label)]
    else:
        for ind, label in enumerate(x_parameter_label_checked):
            x[stel.names.index(label)] = opt.x[ind]
    stel.set_dofs(x)
    
    # Plot the residual history if verbose
    if verbose:
        print(opt.x)
        plt.plot(res_history)
        plt.xlabel('Nb. Iterations')
        plt.ylabel('Objective function')
        plt.show()
    
    # Returns the residual in the optimisation
    return opt.fun



"""
This module contains a function for Newton's method.
"""

import logging
import numpy as np

#logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

def newton(f, x0, jac, niter=20, tol=1e-13, nlinesearch=8):
    """
    Solve a system of nonlinear equations using Newton's method with a
    line search.

    f = function providing the residual vector.
    x0 = initial guess
    jac = function providing the Jacobian.
    niter = max number of Newton iterations.
    tol = stop when the residual norm is less than this.
    """

    x = np.copy(x0)
    x_best = np.copy(x0)
    residual = f(x0)
    initial_residual_norm = np.sqrt(np.sum(residual * residual))
    residual_norm = initial_residual_norm
    logger.info('Beginning Newton method. residual {}'.format(residual_norm))

    newton_tolerance_achieved = False
    for jnewton in range(niter):
        last_residual_norm = residual_norm
        if residual_norm < tol:
            newton_tolerance_achieved = True
            break

        j = jac(x)
        x0 = np.copy(x)
        logger.info('Newton iteration {}'.format(jnewton))
        step_direction = -np.linalg.solve(j, residual)

        step_scale = 1.0
        for jlinesearch in range(nlinesearch):
            x = x0 + step_scale * step_direction
            residual = f(x)
            residual_norm = np.sqrt(np.sum(residual * residual))
            logger.info('  Line search step {} residual {}'.format(jlinesearch, residual_norm))
            if residual_norm < last_residual_norm:
                x_best = np.copy(x)
                break

            step_scale /= 2
            
        if residual_norm >= last_residual_norm:
            logger.info('Line search failed to reduce residual')
            break
    return x_best



#!/usr/bin/env python3

from qsc.util import to_Fourier
import unittest
import os
from scipy.io import netcdf
import numpy as np
import logging
from qsc.qsc import Qsc
from mpi4py import MPI
import vmec

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

def compare_to_vmec(name, r=0.005, nphi=151):
    """
    Check that VMEC can run the input file outputed by pyQSC
    and check that the resulting VMEC output file has
    the expected parameters
    """
    # Add the directory of this file to the specified filename:
    inputFile="input."+str(name).replace(" ","")
    abs_filename = os.path.join(os.path.dirname(__file__), inputFile)
    # Run pyQsc and create a VMEC input file
    logger.info('Creating pyQSC configuration')
    order = 'r2' if name[1] == '2' else 'r1'
    py = Qsc.from_paper(name, nphi=nphi, order=order)
    logger.info('Outputing to VMEC')
    py.to_vmec(inputFile,r)
    # Run VMEC
    fcomm = MPI.COMM_WORLD.py2f()
    logger.info("Calling runvmec. comm={}".format(fcomm))
    ictrl=np.array([15,0,0,0,0], dtype=np.int32)
    vmec.runvmec(ictrl, inputFile, True, fcomm, '')
    # Check that VMEC converged
    assert ictrl[1] == 11
    # Open VMEC output file
    woutFile="wout_"+str(name).replace(" ","")+".nc"
    f = netcdf.netcdf_file(woutFile, 'r')
    # Compare the results
    logger.info('pyQSC iota on axis = '+str(py.iota))
    logger.info('VMEC iota on axis = '+str(-f.variables['iotaf'][()][0]))
    logger.info('pyQSC field on axis = '+str(py.B0))
    logger.info('VMEC bmnc[1][0] = '+str(f.variables['bmnc'][()][1][0]))
    assert np.isclose(py.iota,-f.variables['iotaf'][()][0],rtol=1e-2)
    assert np.isclose(py.B0[0],f.variables['bmnc'][()][1][0],rtol=1e-2)
    vmec.cleanup(True)
    f.close()

def Fourier_Inverse(name, r = 0.05, ntheta = 26, nphi = 51, mpol = 13, ntor = 25, atol=1e-9, rtol=1e-9):
    """
    Compute the Fourier transform of a boundary surface and then
    inverse Fourier transform it to find that it arrives
    at the same surface
    """
    logger.info('Creating pyQSC configuration')
    py = Qsc.from_paper(name, nphi=nphi)

    logger.info('Calculating old R_2D and Z_2D')
    R_2D, Z_2D, phi0_2D = py.Frenet_to_cylindrical(r, ntheta)

    logger.info('Calculating corresponding RBC, RBS, ZBC, ZBS')
    RBC, RBS, ZBC, ZBS = to_Fourier(R_2D, Z_2D, py.nfp, mpol, ntor, py.lasym)
    if not py.lasym:
        RBS = np.zeros((int(2*ntor+1),int(mpol+1)))
        ZBC = np.zeros((int(2*ntor+1),int(mpol+1)))

    logger.info('Inverse Fourier transform')
    nphi_conversion = py.nphi
    theta = np.linspace(0,2*np.pi,ntheta,endpoint=False)
    phi_conversion = np.linspace(0,2*np.pi/py.nfp,nphi_conversion,endpoint=False)
    R_2Dnew = np.zeros((ntheta,nphi_conversion))
    Z_2Dnew = np.zeros((ntheta,nphi_conversion))
    phi2d, theta2d = np.meshgrid(phi_conversion, theta)
    for m in range(mpol+1):
        for n in range(-ntor, ntor+1):
            angle = m * theta2d - n * py.nfp * phi2d
            R_2Dnew += RBC[n+ntor,m] * np.cos(angle) + RBS[n+ntor,m] * np.sin(angle)
            Z_2Dnew += ZBC[n+ntor,m] * np.cos(angle) + ZBS[n+ntor,m] * np.sin(angle)

    logger.info('Check the old and the new match')
    np.testing.assert_allclose(R_2D, R_2Dnew, atol=atol, rtol=rtol)
    np.testing.assert_allclose(Z_2D, Z_2Dnew, atol=atol, rtol=rtol)


class ToVmecTests(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(ToVmecTests, self).__init__(*args, **kwargs)
        logger = logging.getLogger('qsc.qsc')
        logger.setLevel(1)
        self.cases=["r1 section 5.1","r1 section 5.2","r1 section 5.3",\
                    "r2 section 5.1","r2 section 5.2","r2 section 5.3","r2 section 5.4","r2 section 5.5"]

    def test_vmec(self):
        """
        Verify that vmec can actually read the generated input files
        and that vmec's Bfield and iota on axis match the predicted values.
        """
        for case in self.cases:
            logger.info('Going through case '+case)
            compare_to_vmec(case)

    def test_Fourier(self):
        """
        Check that transforming with to_Fourier and then un-transforming gives the identity,
        for both even and odd ntheta and phi, and for lasym True or False.
        """
        for case in self.cases:
            logger.info('Going through case '+case)
            Fourier_Inverse(case, atol=1e-9, rtol=1e-9)
